from __future__ import annotations

from base64 import urlsafe_b64encode as b64e
from dataclasses import dataclass
from typing import Any

from crypto_gltf.data.types import BaseKey, BaseParams
from pydantic import BaseModel, model_validator


class AdaptiveCipherParamsV2(BaseParams):
    # p, q, r here represents the byte to be taken from the float32
    p: int
    q: int
    r: int

    @model_validator(mode="before")
    @classmethod
    def validate_fields(cls, data: Any):
        p, q, r = data.get("p"), data.get("q"), data.get("r")
        assert (
            (p is not None) and (q is not None) and (r is not None)
        )  # all fields must be provided
        assert (
            3 >= p > q > r >= 0
        )  # r must be least significant byte of the 3, p the most significant

        return data


class SBlocks(BaseModel):
    s1: bytes
    s2: bytes
    s3: bytes


class Key(BaseKey):
    k1: bytes | None = None
    k2: bytes | None = None
    k3: bytes | None = None

    @model_validator(mode="before")
    @classmethod
    def validate_fields(cls, data: Any):
        k1, k2, k3 = data.get("k1"), data.get("k2"), data.get("k3")
        assert k1 or k2 or k3
        return data

    def __str__(self) -> str:
        return str({k: b64e(v) for k, v in self.__dict__.items() if v is not None})

    @property
    def filled(self) -> bool:
        """Returns true if all subkeys below highest non-null subkey exist"""
        if self.k3 is not None:
            return (self.k2 is not None) and (self.k1 is not None)
        elif self.k2 is not None:
            return self.k1 is not None
        else:
            return self.k1 is not None  # could remove but kept for clarity

    @property
    def size(self) -> int:
        """Returns the number of non null subkeys"""
        return len([ki for ki in self.__dict__.values() if ki is not None])


class BlockSelection(BaseModel):
    p: bool = False
    q: bool = False
    r: bool = False

    @classmethod
    def all(cls) -> BlockSelection:
        """Select all blocks"""
        return cls(p=True, q=True, r=True)



from __future__ import annotations

import binascii
import secrets
import time
from base64 import urlsafe_b64decode as b64d
from base64 import urlsafe_b64encode as b64e

from crypto_gltf.data.types import EncryptionResponse
from cryptography.fernet import InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from pydantic import BaseModel

backend = default_backend()


class AAD(BaseModel):
    """Additional associated data"""

    timestamp: bytes
    iv: bytes
    tag: bytes

    @classmethod
    def b64d(cls, aad_b64: bytes, block_size: int) -> AAD:
        try:
            data = b64d(aad_b64)
        except (TypeError, binascii.Error):
            raise InvalidToken
        timestamp, iv, tag = data[:8], data[8 : block_size // 8 + 8], data[-16:]
        return cls(timestamp=timestamp, iv=iv, tag=tag)

    @property
    def b64e(self) -> bytes:
        """URL and filesystem safe b64 format"""
        return b64e(self.timestamp + self.iv + self.tag)


def aes_gcm_encrypt(
    message: bytes, key: bytes
) -> EncryptionResponse[bytes, bytes, bytes]:
    """encrypt message using AES-GCM cipher"""

    current_time = int(time.time()).to_bytes(8, "big")
    algorithm = algorithms.AES(key)
    iv = secrets.token_bytes(algorithm.block_size // 8)
    cipher = Cipher(algorithm, modes.GCM(iv), backend=backend)
    encryptor = cipher.encryptor()
    encryptor.authenticate_additional_data(current_time)

    ciphertext = encryptor.update(message) + encryptor.finalize()
    aad = b64e(current_time + iv + encryptor.tag)

    return EncryptionResponse[bytes, bytes, bytes](
        ciphertext=ciphertext, aad=aad, key=key
    )


def aes_gcm_decrypt(ciphertext: bytes, aad_b64: bytes, key: bytes, ttl=None) -> bytes:
    """decrypt token using AES-GCM cipher"""

    algorithm = algorithms.AES(key)
    aad = AAD.b64d(aad_b64=aad_b64, block_size=algorithm.block_size)

    if ttl is not None:
        current_time = int(time.time())
        time_encrypted = int.from_bytes(aad.timestamp, "big")
        if time_encrypted + ttl < current_time or current_time + 60 < time_encrypted:
            # too old or created well before our current time + 1 h to account for clock skew
            raise InvalidToken

    cipher = Cipher(algorithm, modes.GCM(aad.iv, aad.tag), backend=backend)
    decryptor = cipher.decryptor()
    decryptor.authenticate_additional_data(aad.timestamp)
    return decryptor.update(ciphertext) + decryptor.finalize()



from __future__ import annotations

import os
import struct

import numpy as np
from crypto_gltf.data.asset_file_data_types import OffData
from crypto_gltf.data.types import AAD_DATA, Composition
from crypto_gltf.io.file.base_file import BaseFile
from crypto_gltf.io.plaintext.plnm import PlnM
from loguru import logger


class OffFile(BaseFile):
    data: OffData

    @classmethod
    def load(cls, import_path: str) -> OffFile:
        """Load an OFF Mesh file"""

        # Parse mesh from OFF file
        import_path_bin = os.fsencode(import_path)
        file = open(import_path_bin, "r")
        first_line = file.readline().rstrip()
        use_colors = first_line == "COFF"
        colors = []

        # handle blank and comment lines after the first line
        line = file.readline()
        while line.isspace() or line[0] == "#":
            line = file.readline()

        vcount, fcount, ecount = [int(x) for x in line.split()]
        verts = []
        faces = []
        edges = []
        i = 0
        while i < vcount:
            line = file.readline()
            if line.isspace():
                continue  # skip empty lines
            try:
                bits = [float(x) for x in line.split()]
                px = bits[0]
                py = bits[1]
                pz = bits[2]
                if use_colors:
                    colors.append(
                        [
                            float(bits[3]) / 255,
                            float(bits[4]) / 255,
                            float(bits[5]) / 255,
                            1.0,
                        ]
                    )

            except ValueError:
                i = i + 1
                continue
            verts.append((px, py, pz))
            i = i + 1

        i = 0
        while i < fcount:
            line = file.readline()
            if line.isspace():
                continue  # skip empty lines
            try:
                splitted = line.split()
                ids = list(map(int, splitted))
                if len(ids) > 3:
                    faces.append(tuple(ids[1:]))
                elif len(ids) == 3:
                    edges.append(tuple(ids[1:]))
            except ValueError:
                i = i + 1
                continue
            i = i + 1

        logger.info(f".off file loaded from {import_path}")

        return cls(
            import_path=import_path,
            data=OffData(
                verts=np.array(verts, dtype=np.float32),
                faces=faces,
                colors=np.array(colors, dtype=np.float32),
            ),
            filename_ext=".off",
        )

    def save(self, export_dir: str, images_encrypted: bool = False) -> str:
        """Save to OFF Mesh file"""

        off_data = self.data

        verts = off_data.verts
        faces = off_data.faces
        colors = off_data.colors

        # Write geometry to file
        export_filepath = f"{export_dir}/{self.export_filename}"
        export_dir_bin = os.fsencode(export_filepath)
        fp = open(export_dir_bin, "w")

        if colors.size > 0:
            fp.write("COFF\n")
        else:
            fp.write("OFF\n")

        fp.write("%d %d 0\n" % (len(verts), len(faces)))

        for i, vert in enumerate(verts):
            fp.write(f"{vert[0]:.64} {vert[1]:.64} {vert[2]:.64}")
            if colors.size > 0:
                fp.write(f" {colors[i][0]:d} {colors[i][1]:d} {colors[i][2]:d} 255")
            fp.write("\n")

        # for face in faces:
        for i, face in enumerate(faces):
            fp.write("%d" % len(face))
            for vid in face:
                fp.write(" %d" % vid)
            fp.write("\n")

        fp.close()

        logger.info(f".off file saved to {export_filepath}")

        return export_filepath

    @property
    def plnm(self) -> PlnM:
        """'Get plaintext from OffFile"""

        verts, colors = (
            self.data.verts,
            self.data.colors,
        )  # skip faces as they're of inhomogeneous shape
        return PlnM(meshes=[verts, colors], images=[], meshes_dim=2, images_dim=0)

    def insert_plnm(self, plnm: PlnM) -> None:
        """Insert plaintext in OffFile
        assumes the plaintext data came from the same OffFile"""

        self.data.verts = plnm.meshes[0]
        self.data.colors = plnm.meshes[1]

    def embed_aad(self, aad: AAD_DATA) -> None:
        """Embed aad data in file for retrieval during decryption"""

        assert Composition.AAD == aad.aad
        num_used_vertices = max(max(x) for x in self.data.faces) + 1

        # so that massive aad vertices don't interfere in rendering, we set all bits larger than 16 to 0
        # and divide the array in half
        aad_16 = np.frombuffer(aad.aad.tobytes(), dtype=np.uint16)
        buffer = bytearray(b"")
        for val in aad_16:
            buffer.extend(struct.pack(">HH", val, np.uint16(0)))

        aad_float32 = np.frombuffer(buffer, dtype=np.float32).reshape((-1, 3))
        params = np.array(
            [
                [
                    aad.meshes_params.get("p"),
                    aad.meshes_params.get("q"),
                    aad.meshes_params.get("r"),
                ]
            ]
        )
        self.data.verts = np.concatenate(
            (self.data.verts[:num_used_vertices], aad_float32, params), dtype=np.float32
        )

    @property
    def aad(self) -> AAD_DATA:
        """Retrieve embedded aad data"""
        num_used_vertices = max(max(x) for x in self.data.faces) + 1
        params = self.data.verts[-1]
        aad = self.data.verts[num_used_vertices:][:-1]
        self.data.verts = self.data.verts[
            :num_used_vertices
        ]  # remove embedded aad data

        # retrieve split aad values
        buffer = bytearray(b"")
        for val in aad.flatten():
            bits, _ = struct.unpack(">HH", val)
            buffer.extend(np.uint16(bits).tobytes())

        aad = np.frombuffer(buffer, dtype=np.uint32).reshape((-1, 3))

        params_dic = {"p": int(params[0]), "q": int(params[1]), "r": int(params[2])}
        return AAD_DATA(
            aad=aad, encrypt_images=False, meshes_params=params_dic, images_params={}
        )



import secrets
from time import time

import numpy as np
from crypto_gltf.data.types import EncryptionResponse
from crypto_gltf.encrypt.adaptive.base import AdaptiveBaseModel
from crypto_gltf.encrypt.adaptive.cryptography.aes_gcm import aes_gcm_encrypt
from crypto_gltf.encrypt.adaptive.types import AdaptiveCipherParams, BufferView, Key
from crypto_gltf.encrypt.adaptive.utils import get_bits, put_bits
from loguru import logger


class AdaptiveEncryptionModel(AdaptiveBaseModel):

    @staticmethod
    def _encrypt(
        data: list[tuple[np.ndarray, AdaptiveCipherParams]], key: Key
    ) -> EncryptionResponse[list[np.ndarray], np.ndarray, Key]:

        tic = time()

        s1, s2, s3 = bytearray(b""), bytearray(b""), bytearray(b"")

        for arr, params in data:

            buffer_view = BufferView.from_shape(shape=arr.shape, params=params)

            p_buffer, q_buffer, r_buffer = (
                buffer_view.p_buffer,
                buffer_view.q_buffer,
                buffer_view.r_buffer,
            )

            get_bits(
                arr,
                p_buffer,
                params.pstart,
                params.pstop,
                secrets.randbits(32),
            )
            get_bits(
                arr,
                q_buffer,
                params.qstart,
                params.qstop,
                secrets.randbits(32),
            )
            get_bits(
                arr,
                r_buffer,
                params.rstart,
                params.rstop,
                secrets.randbits(32),
            )

            s1.extend(p_buffer.tobytes())
            s2.extend(q_buffer.tobytes())
            s3.extend(r_buffer.tobytes())

        toc = time() - tic
        assert key.k1 and key.k2 and key.k3

        r1 = aes_gcm_encrypt(message=s1, key=key.k1)
        r2 = aes_gcm_encrypt(message=s2, key=key.k2)
        r3 = aes_gcm_encrypt(message=s3, key=key.k3)

        # combine aads
        aad = bytearray(b"")
        aad.extend(r1.aad)
        aad.extend(r2.aad)
        aad.extend(r3.aad)

        tic = time()

        r1_arr = np.frombuffer(r1.ciphertext, dtype=np.uint32)
        r2_arr = np.frombuffer(r2.ciphertext, dtype=np.uint32)
        r3_arr = np.frombuffer(r3.ciphertext, dtype=np.uint32)

        padding = []
        poffset, qoffset, roffset = 0, 0, 0
        for arr, params in data:

            buffer_view = BufferView.from_shape(shape=arr.shape, params=params)
            put_bits(
                arr,
                r1_arr[poffset : poffset + buffer_view.pbufflen],
                params.pstart,
                params.pstop,
            )
            put_bits(
                arr,
                r2_arr[qoffset : qoffset + buffer_view.qbufflen],
                params.qstart,
                params.qstop,
            )
            put_bits(
                arr,
                r3_arr[roffset : roffset + buffer_view.rbufflen],
                params.rstart,
                params.rstop,
            )

            padding.append(
                [
                    r1_arr[poffset + buffer_view.pbufflen - 1],
                    r2_arr[qoffset + buffer_view.qbufflen - 1],
                    r3_arr[roffset + buffer_view.rbufflen - 1],
                ]
            )
            poffset += buffer_view.pbufflen
            qoffset += buffer_view.qbufflen
            roffset += buffer_view.rbufflen

        if not poffset == len(r1_arr):
            raise Exception(f"Entire p buffer has not been inserted")
        if not qoffset == len(r2_arr):
            raise Exception(f"Entire q buffer has not been inserted")
        if not roffset == len(r3_arr):
            raise Exception(f"Entire r buffer has not been inserted")

        aad_arr = np.column_stack(
            (
                np.frombuffer(r1.aad, dtype=np.uint32),
                np.frombuffer(r2.aad, dtype=np.uint32),
                np.frombuffer(r3.aad, dtype=np.uint32),
            )
        )  # convert to arr so we can embed in file
        aad = np.concatenate(
            (np.array(padding, dtype=np.uint32), aad_arr)
        )  # add padding to authentification information

        logger.debug(
            f"Byte retrieval/insertion and reshaping took {time()-tic + toc} seconds"
        )

        return EncryptionResponse[list[np.ndarray], np.ndarray, Key](
            ciphertext=[item[0] for item in data],
            aad=aad,
            key=key,
        )



from typing import Literal

import numpy as np
from crypto_gltf.utils.bit_utils import bit_slice

stack_dict = {"HEIGHT": 0, "WIDTH": 1, "DEPTH": 2}


def split_array_stack(
    stacked_data: np.ndarray,
    stacktype: Literal["HEIGHT", "WIDTH", "DEPTH"],
) -> list[np.ndarray]:
    """split numpy array stack by height, width or depth"""

    dim = stacked_data.shape[stack_dict[stacktype]]

    splits = np.split(
        ary=stacked_data,
        indices_or_sections=dim,
        axis=stack_dict[stacktype],
    )
    return [arr.reshape(arr.shape[0], arr.shape[1]) for arr in splits]


def array_bit_slice(arr: np.ndarray, start: int, stop: int) -> np.ndarray:
    """Returns array of bit slices for each element of arr"""
    assert len(arr.shape) == 1  # make sure array is 1D
    return np.array([bit_slice(int(val), start, stop) for val in arr])


def array_bit_slice_eq(
    arr1: np.ndarray, arr2: np.ndarray, start: int, stop: int
) -> bool:
    """Returns equality of bit slices of every element in arr1 nad arr2"""
    arr1_slice = array_bit_slice(arr1.flatten(), start, stop)
    arr2_slice = array_bit_slice(arr2.flatten(), start, stop)
    return np.array_equal(arr1_slice, arr2_slice)



import secrets

import numpy as np
from crypto_gltf.encrypt.deprecit.adaptive_v2.types import (
    AdaptiveCipherParamsV2,
    BlockSelection,
    Key,
)
from crypto_gltf.encrypt.deprecit.adaptive_v2.utils import combine_sblocks, get_sblocks
from crypto_gltf.io.plaintext.plnm import PlnM
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from loguru import logger


def aes_sha(sblock: bytes, ki: bytes) -> bytes:
    """Generate subsequent subkey k_(i-1) using AES followed by sha"""

    assert len(sblock) == 32

    cipher = Cipher(algorithms.AES(ki), modes.ECB(), default_backend())
    encryptor = cipher.encryptor()
    ciphertext = encryptor.update(sblock) + encryptor.finalize()

    digest = hashes.Hash(hashes.SHA256())
    digest.update(sblock)
    digest.update(ciphertext)

    return digest.finalize()


def get_source_arrays(float_arrs: list[np.ndarray]) -> list[np.ndarray]:
    """Return first n matrices in float_arrs such that the sum of matrix sizes is greater than 32.
    (32*4 = 256 = AES keylength)"""

    source_arrays: list[np.ndarray] = []

    for arr in float_arrs:
        source_arrays.append(arr)
        size = sum([arr.size for arr in source_arrays])
        if size > 32:
            break

    if sum([arr.size for arr in source_arrays]) < 32:
        raise Exception("Mesh too small to generate keys")

    return source_arrays


def generate_keys(
    plnm: PlnM, params: AdaptiveCipherParamsV2, k3: bytes | None = None
) -> Key:
    """Calculates k1, k2 and k3 for a given plaintext asset"""

    float_arrs = [mesh for mesh in plnm.meshes if mesh.dtype.kind == "f"]
    source_arrays = get_source_arrays(float_arrs=float_arrs)

    sblocks_list = []
    for arr in source_arrays:
        sblocks_list.append(
            get_sblocks(
                arr=arr,
                params=params,
                to_get=BlockSelection.all(),
            )
        )
    combined_sblocks = combine_sblocks(sblocks_list=sblocks_list)

    s3 = combined_sblocks.s3[:32]
    s2 = combined_sblocks.s2[:32]

    if k3:
        assert len(k3) == 32
        logger.info("User specified key detected, using key provided.")
    else:
        k3 = secrets.token_bytes(32)

    k2 = aes_sha(sblock=s3, ki=k3)
    k1 = aes_sha(sblock=s2, ki=k2)

    return Key(k1=k1, k2=k2, k3=k3)


def get_k2(
    float_arrs: list[np.ndarray], params: AdaptiveCipherParamsV2, k3: bytes
) -> bytes:
    """Calculate k2 from k3 for a given plaintext"""

    source_arrays = get_source_arrays(float_arrs=float_arrs)

    sblocks_list = []
    for arr in source_arrays:
        sblocks_list.append(
            get_sblocks(arr=arr, params=params, to_get=BlockSelection(r=True))
        )

    combined_sblocks = combine_sblocks(sblocks_list=sblocks_list)

    s3 = combined_sblocks.s3[:32]
    k2 = aes_sha(sblock=s3, ki=k3)

    return k2


def get_k1(
    float_arrs: list[np.ndarray], params: AdaptiveCipherParamsV2, k2: bytes
) -> bytes:
    """Calculate k1 from k2 for a given plaintext"""

    source_arrays = get_source_arrays(float_arrs)

    sblocks_list = []
    for arr in source_arrays:
        sblocks_list.append(
            get_sblocks(arr=arr, params=params, to_get=BlockSelection(q=True))
        )

    combined_sblocks = combine_sblocks(sblocks_list=sblocks_list)

    s2 = combined_sblocks.s2[:32]
    k1 = aes_sha(sblock=s2, ki=k2)

    return k1



from __future__ import annotations

from time import time

import numpy as np
from crypto_gltf.data.types import AAD_DATA, EncryptionResponse
from crypto_gltf.encrypt.adaptive.cryptography.key_gen import generate_keys, get_subkey
from crypto_gltf.encrypt.adaptive.decrypt import AdaptiveDecryptionModel
from crypto_gltf.encrypt.adaptive.encrypt import AdaptiveEncryptionModel
from crypto_gltf.encrypt.adaptive.types import (
    AdaptiveCipherParams,
    BlockSelection,
    ImagesAdaptiveCipherParams,
    Key,
    MeshesAdaptiveCipherParams,
)
from crypto_gltf.io.plaintext.plnm import PlnM
from loguru import logger


class AdaptiveCryptoSystemV3(AdaptiveEncryptionModel, AdaptiveDecryptionModel):
    """Adaptive cipher system"""

    AAD: bool = True

    @staticmethod
    def encrypt(
        plnm: PlnM,
        meshes_cipher_params: MeshesAdaptiveCipherParams,
        images_cipher_params: ImagesAdaptiveCipherParams,
        k3: bytes | None = None,
        encrypt_images: bool = False,
    ) -> EncryptionResponse[PlnM, AAD_DATA, Key]:

        tic = time()

        # Get non-integer type mesh arrays
        # i.e. vertices and colors
        meshes_float_arrs_idxs = np.array(
            [
                idx
                for idx, mesh in enumerate(plnm.meshes)
                if mesh.dtype.kind == "f" and mesh.size > 0
            ]
        )
        float_arrs = [
            plnm.meshes[i].astype(np.float32) for i in meshes_float_arrs_idxs
        ]  # convert to float32

        if len(float_arrs) == 0:
            raise Exception(f"No asset data to encrypt.")
        if not k3:
            key = generate_keys(float_arrs, meshes_cipher_params)
        else:
            # a key has been provided
            if not len(k3) == 32:
                raise Exception(f"Key must be 32 bytes long, received {len(k3)} bytes")
            key = generate_keys(float_arrs[:50], meshes_cipher_params, k3)

        data: list[tuple[np.ndarray, AdaptiveCipherParams]] = [
            (float_arr, meshes_cipher_params) for float_arr in float_arrs
        ]
        if encrypt_images:
            data += [(img, images_cipher_params) for img in plnm.images]

        encryption_response = AdaptiveEncryptionModel._encrypt(data=data, key=key)
        for idx, i in enumerate(meshes_float_arrs_idxs):
            plnm.meshes[i] = encryption_response.ciphertext[idx]

        if encrypt_images:
            plnm.images = encryption_response.ciphertext[len(float_arrs) :]

        aad = AAD_DATA(
            aad=encryption_response.aad,
            encrypt_images=encrypt_images,
            meshes_params=meshes_cipher_params.__dict__,
            images_params=images_cipher_params.__dict__,
        )

        logger.debug(f"Adaptive mesh encryption took {time()-tic} seconds.")
        logger.success(f"Mesh encrypted")

        return EncryptionResponse[PlnM, AAD_DATA, Key](
            ciphertext=plnm,
            aad=aad,
            key=key,
        )

    @staticmethod
    def decrypt(
        plnm: PlnM,
        key: Key,
        aad: AAD_DATA,
    ) -> PlnM:
        tic = time()

        meshes_cipher_params = MeshesAdaptiveCipherParams(**aad.meshes_params)

        if key.size == 0:
            raise Exception(f"Please provide a key")

        # Get non-integer type arrays
        # i.e. vertices and colors
        meshes_float_arrs_idxs = np.array(
            [
                idx
                for idx, mesh in enumerate(plnm.meshes)
                if mesh.dtype.kind == "f" and mesh.size > 0
            ]
        )

        float_arrs = [plnm.meshes[i] for i in meshes_float_arrs_idxs]

        data: list[tuple[np.ndarray, AdaptiveCipherParams]] = [
            (float_arr, meshes_cipher_params) for float_arr in float_arrs
        ]
        if aad.encrypt_images:
            images_cipher_params = ImagesAdaptiveCipherParams(**aad.images_params)
            data += [(img, images_cipher_params) for img in plnm.images]

        if key.k3:
            data = AdaptiveDecryptionModel._decrypt(
                data=data, key=key, aad=aad.aad, selection=BlockSelection(r=True)
            )

            key.k2 = get_subkey(
                plnm_arrs=[item[0] for item in data[: len(float_arrs)]],
                params=meshes_cipher_params,
                ki=key.k3,
                selection=BlockSelection(r=True),
            )

        if key.k2:
            data = AdaptiveDecryptionModel._decrypt(
                data=data, key=key, aad=aad.aad, selection=BlockSelection(q=True)
            )

            key.k1 = get_subkey(
                plnm_arrs=[item[0] for item in data[: len(float_arrs)]],
                params=meshes_cipher_params,
                ki=key.k2,
                selection=BlockSelection(q=True),
            )

        assert key.k1 is not None

        decrypted_data = AdaptiveDecryptionModel._decrypt(
            data=data, key=key, aad=aad.aad, selection=BlockSelection(p=True)
        )

        for idx, i in enumerate(meshes_float_arrs_idxs):
            plnm.meshes[i] = decrypted_data[idx][0]

        assert key.filled

        logger.debug(f"Adaptive mesh decryption took {time()-tic} seconds.")
        logger.success(f"Mesh decrypted")

        return plnm



from typing import Literal

import numpy as np


def get_bytes(arr: np.ndarray, position: Literal[0, 1, 2, 3]) -> bytes:
    """Get byte at position for every element in float32 arr"""

    if arr.dtype != np.float32:
        raise Exception(f"Invalid array dtype {arr.dtype}, must be float32")

    bytes_arr = np.ndarray.tobytes(arr)
    buffer = bytearray(b"")
    for i in range(arr.size):
        buffer.extend(bytes_arr[4 * i + position : 4 * i + position + 1])

    return buffer


def insert_bytes(
    arr: np.ndarray, position: Literal[0, 1, 2, 3], buffer: bytes
) -> np.ndarray:
    """Insert bytes at position in every element in float32 arr"""

    bytes_arr = np.ndarray.tobytes(arr)
    for i in range(arr.size):
        bytes_arr = (
            bytes_arr[: 4 * i + position]
            + buffer[i : i + 1]
            + bytes_arr[4 * i + position + 1 :]
        )
    return np.frombuffer(bytes_arr, dtype=np.float32)



from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from crypto_gltf.data.types import PlnMDataType
from crypto_gltf.io.plaintext.base_plaintext import BasePlainText
from PIL import Image


@dataclass
class PlnM(BasePlainText):
    """Plaintext data class for multi matrix 3D asset attributes (colors, meshes etc.)"""

    meshes: list[np.ndarray]
    images: list[np.ndarray]

    def __eq__(self, plnm: PlnM) -> bool:
        if (self.meshes_dim != plnm.meshes_dim) or (self.images_dim != plnm.images_dim):
            return False
        meshes_eq = all(
            [np.allclose(arr1, arr2) for arr1, arr2 in zip(self.meshes, plnm.meshes)]
        )  # allclose to account for type conversions to np.float32
        images_eq = all(
            [np.array_equal(arr1, arr2) for arr1, arr2 in zip(self.images, plnm.images)]
        )
        return meshes_eq and images_eq

    def __copy__(self):
        return PlnM(
            meshes=[np.copy(arr) for arr in self.meshes],
            images=[np.copy(arr) for arr in self.images],
            meshes_dim=self.meshes_dim,
            images_dim=self.images_dim,
        )

    @property
    def renders(self) -> list[Image.Image]:
        """Render image data classes"""

        renders = []
        for image in self.images:
            datatype = PlnMDataType.from_data(image)
            assert datatype.renderable
            renders.append(Image.fromarray(image, datatype.typed_name))

        return renders

    @classmethod
    def from_images(cls, images: list[Image.Image]) -> PlnM:
        """'Get plaintext from list of images"""

        img_arrays = [np.asarray(image) for image in images]
        return cls(meshes=[], images=img_arrays, meshes_dim=0, images_dim=len(images))



from __future__ import annotations

import numpy as np
from crypto_gltf.encrypt.deprecit.ca.base import CACipherParams
from crypto_gltf.encrypt.deprecit.ca.ca import ExpConwaysHCA
from crypto_gltf.encrypt.deprecit.ca.decrypt import CADecryptionModel
from crypto_gltf.encrypt.deprecit.ca.encrypt import CAEncryptionModel
from crypto_gltf.encrypt.deprecit.ca.params import RANDOM_VARIANCE
from crypto_gltf.io.plaintext.plnm import PlnM
from loguru import logger
from pydantic import BaseModel, ConfigDict


class CACryptoSystem(BaseModel):
    """Conways game of life cellular automata cipher system"""

    AAD: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True)
    plnm: PlnM  # Data

    meshes_expHCA: ExpConwaysHCA
    images_expHCA: ExpConwaysHCA

    meshes_cipher_params: CACipherParams
    images_cipher_params: CACipherParams

    generations_computed: bool = False

    @classmethod
    def from_pre_encryption(
        cls,
        plnm: PlnM,
        meshes_cipher_params: CACipherParams,
        images_cipher_params: CACipherParams,
        random_seed: int = 0,
    ) -> CACryptoSystem:
        """Initialise class using pre-encryption steps"""
        if random_seed:
            logger.info(f"Setting random seed = {random_seed}")
            np.random.seed(random_seed)

        meshes_cipher_params.validate_idl_points(plnm.meshes_dim)
        images_cipher_params.validate_idl_points(plnm.images_dim)

        meshes_expHCA = ExpConwaysHCA(
            dimension=CACryptoSystem.get_expHCA_dimension(plnm.meshes)
        )
        images_expHCA = ExpConwaysHCA(
            dimension=CACryptoSystem.get_expHCA_dimension(plnm.images)
        )

        return cls(
            plnm=plnm,
            meshes_expHCA=meshes_expHCA,
            images_expHCA=images_expHCA,
            meshes_cipher_params=meshes_cipher_params,
            images_cipher_params=images_cipher_params,
        )

    @staticmethod
    def get_expHCA_dimension(data: list[np.ndarray]) -> tuple[int, int]:
        """Get expHCA dimensions from data and random variance"""
        if len(data) == 0:
            return (0, 0)
        mat_shapes = np.zeros((len(data), 3), dtype=int)
        for idx, mat in enumerate(data):
            shape = mat.shape
            mat_shapes[idx][: len(shape)] = shape

        r, c = max(mat_shapes[..., 0]), max(
            mat_shapes[..., 1]
        )  # max row and column numbers in list
        n1, n2 = np.random.randint(0, RANDOM_VARIANCE, 2)  # ExpCA expansion values
        dimension = (r + n1, c + n2)
        logger.info(f"Using CA dimension {dimension}")
        return dimension

    def compute_generations(self) -> None:
        """Compute expHCA generations for mesh and image automatas"""
        self.meshes_expHCA = CAEncryptionModel.compute_gens(
            expHCA=self.meshes_expHCA, cipher_params=self.meshes_cipher_params
        )
        self.images_expHCA = CAEncryptionModel.compute_gens(
            expHCA=self.images_expHCA, cipher_params=self.images_cipher_params
        )
        self.generations_computed = True

    def encrypt(self) -> None:
        """Encrypt all meshes and images"""
        if not self.generations_computed:
            raise Exception("Generations not yet computed")
        self.plnm.meshes = CAEncryptionModel.encrypt(
            plain_text=self.plnm.meshes,
            expHCA=self.meshes_expHCA,
            cipher_params=self.meshes_cipher_params,
        )
        self.plnm.images = CAEncryptionModel.encrypt(
            plain_text=self.plnm.images,
            expHCA=self.images_expHCA,
            cipher_params=self.images_cipher_params,
        )

    def decrypt(self) -> None:
        """Decrypt all meshes and images"""
        if not self.generations_computed:
            raise Exception("Generations not yet computed")
        self.plnm.meshes = CADecryptionModel.decrypt(
            cipher_text=self.plnm.meshes,
            expHCA=self.meshes_expHCA,
            cipher_params=self.meshes_cipher_params,
        )
        self.plnm.images = CADecryptionModel.decrypt(
            cipher_text=self.plnm.images,
            expHCA=self.images_expHCA,
            cipher_params=self.images_cipher_params,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Type

import numpy as np
from crypto_gltf.data.types import AAD_DATA, BaseKey, BaseParams, EncryptionResponse
from crypto_gltf.encrypt.adaptive.system import AdaptiveCryptoSystemV3
from crypto_gltf.encrypt.adaptive.types import (
    ImagesAdaptiveCipherParams,
    Key,
    MeshesAdaptiveCipherParams,
)
from crypto_gltf.io.file.base_file import BaseFile
from crypto_gltf.io.file.file import File
from crypto_gltf.io.plaintext.plnm import PlnM


@dataclass
class Asset:
    file: BaseFile
    images_encrypted: bool = False

    @classmethod
    def load(
        cls,
        import_filepath: str,
    ) -> Asset:
        """Load file from import_filepath"""
        return cls(file=File(import_filepath=import_filepath))

    def save(self, export_dir: str) -> str:
        """Save file to export_dir"""
        return self.file.save(
            export_dir=export_dir, images_encrypted=self.images_encrypted
        )

    def encrypt(
        self,
        meshes_cipher_params: tuple[int, int, int] = (2, 2, 10),
        images_cipher_params: tuple[int, int, int] = (1, 1, 6),
        key: bytes | None = None,
        encrypt_images: bool = False,
    ) -> EncryptionResponse[PlnM, AAD_DATA, Key]:
        """Encrypt a file"""
        self.images_encrypted = encrypt_images

        mesh_params = MeshesAdaptiveCipherParams(
            p=meshes_cipher_params[0],
            q=meshes_cipher_params[1],
            r=meshes_cipher_params[2],
        )
        img_params = ImagesAdaptiveCipherParams(
            p=images_cipher_params[0],
            q=images_cipher_params[1],
            r=images_cipher_params[2],
        )
        response = AdaptiveCryptoSystemV3.encrypt(
            plnm=self.file.plnm,
            meshes_cipher_params=mesh_params,
            images_cipher_params=img_params,
            k3=key,
            encrypt_images=encrypt_images,
        )
        self.file.insert_plnm(response.ciphertext)

        assert response.aad is not None
        self.file.embed_aad(response.aad)

        return response

    def decrypt(
        self,
        k1: bytes | None = None,
        k2: bytes | None = None,
        k3: bytes | None = None,
    ) -> bool:
        """Decrypt a file"""
        key = Key(k1=k1, k2=k2, k3=k3)
        aad = (
            self.file.aad
        )  # important we do this before extracting plnm, else aad extracted as well

        plnm = AdaptiveCryptoSystemV3.decrypt(plnm=self.file.plnm, key=key, aad=aad)
        self.file.insert_plnm(plnm)
        return True



from __future__ import annotations

import os
from functools import cached_property
from tempfile import TemporaryDirectory

from crypto_gltf.data.asset_file_data_types import AssetFileDataType
from crypto_gltf.data.types import AAD_DATA, Extension
from crypto_gltf.io.plaintext.plnm import PlnM
from pydantic import BaseModel, ConfigDict


class BaseFile(BaseModel):
    """Base class for 3D asset file types"""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    import_path: str
    filename_ext: Extension.ASSET
    data: AssetFileDataType

    @classmethod
    def load(cls, import_path: str) -> BaseFile:
        raise NotImplementedError()

    def save(self, export_dir: str, images_encrypted: bool = False) -> str:
        """Returns export filepath"""
        raise NotImplementedError()

    @cached_property
    def basename(self) -> str:
        return os.path.basename(self.import_path)

    @cached_property
    def filename(self) -> str:
        return os.path.splitext(self.basename)[0]

    @cached_property
    def import_dirpath(self) -> str:
        return os.path.dirname(self.import_path)

    @cached_property
    def import_dirname(self) -> str:
        return os.path.basename(self.import_dirpath)

    @cached_property
    def export_filename(self) -> str:
        return f"exported_{self.filename}{self.filename_ext}"

    @property
    def plnm(self) -> PlnM:
        """Return plaintext data from file"""
        raise NotImplementedError()

    def insert_plnm(self, plnm: PlnM) -> None:
        """Insert plaintext data in file
        assumes the plaintext data came from the same file"""
        raise NotImplementedError()

    def embed_aad(self, aad: AAD_DATA) -> None:
        """Embed aad data in file for retrieval during decryption"""
        raise NotImplementedError()

    @property
    def aad(self) -> AAD_DATA:
        """Retrieve embedded aad data"""
        raise NotImplementedError()



from dataclasses import dataclass
from typing import Any

import numpy as np
from crypto_gltf.data.types import Composition, JSONDict
from PIL import Image
from pydantic import BaseModel, ConfigDict, model_validator


@dataclass
class OffData:
    model_config = ConfigDict(arbitrary_types_allowed=True)
    verts: type[np.ndarray]
    faces: list[Any]
    colors: type[np.ndarray]

    @model_validator(mode="before")
    @classmethod
    def validate_fields(cls, data):
        verts = data.get("verts")
        assert Composition.VERTS == verts
        return data


@dataclass
class Gltf2Data:
    gltf: JSONDict
    images: dict[int, Image.Image]
    accessors: dict[int, np.ndarray]
    is_glb: bool


class AssetFileDataType:
    OFF = OffData
    GLTF2 = Gltf2Data



import json
import os
import struct
from functools import cached_property

from crypto_gltf.data.asset_file_data_types import Gltf2Data
from crypto_gltf.io.file.gltf2.com.gltf2_io import Buffer, BufferView
from crypto_gltf.io.file.gltf2.exp.gltf2_io_binary_data import (
    BinaryData as ExpBinaryData,
)
from crypto_gltf.io.file.gltf2.exp.gltf2_io_buffer import Buffer as ExpBuffer
from crypto_gltf.io.file.gltf2.utils import uri_to_path


class GlTF2Exporter:
    """Class to handle glTF2 export"""

    def __init__(
        self,
        export_dir: str,
        filename: str,
        is_glb: bool,
        filename_ext: str,
        images_encrypted: bool = False,
    ):
        self.export_buffer: ExpBuffer = ExpBuffer()
        self.buffer_views: list[BufferView] = []
        self.export_dir: str = export_dir
        self.filename: str = filename
        self.is_glb: bool = is_glb
        self.filename_ext: str = filename_ext
        self.images_encrypted: bool = images_encrypted

    @cached_property
    def export_filepath(self) -> str:
        return f"{self.export_dir}/{self.filename}{self.filename_ext}"

    def export(self, gltf_data: Gltf2Data) -> str:
        """Export glTF2 file"""

        self.gltf = gltf_data.gltf.copy()

        # Export accessors
        for i in range(len(gltf_data.accessors)):
            self.export_accessor(gltf_data, i)

        # Export images
        for i in range(len(gltf_data.images)):
            self.export_image(gltf_data, i)

        if self.is_glb:
            uri = None
        else:
            uri = "scene.bin"

        buffer = Buffer(
            byte_length=self.export_buffer.byte_length,
            extensions=None,
            extras=None,
            name=None,
            uri=uri,
        )

        # Set buffer and buffer views
        self.gltf["bufferViews"] = [
            buffer_view.to_dict() for buffer_view in self.buffer_views
        ]
        self.gltf["buffers"] = [buffer.to_dict()]

        gltf_bytes = json.dumps(self.gltf).encode("utf-8")
        buffer_bytes = self.export_buffer.to_bytes()

        match self.is_glb:
            case True:
                with open(self.export_filepath, "wb") as f:

                    length_gltf = len(gltf_bytes)
                    spaces_gltf = (4 - (length_gltf & 3)) & 3
                    length_gltf += spaces_gltf

                    length_bin = len(buffer_bytes)
                    zeros_bin = (4 - (length_bin & 3)) & 3
                    length_bin += zeros_bin

                    length = 12 + 8 + length_gltf
                    if length_bin > 0:
                        length += 8 + length_bin

                    # Header (Version 2)
                    f.write("glTF".encode())
                    f.write(struct.pack("I", 2))
                    f.write(struct.pack("I", length))

                    # Chunk 0 (JSON)
                    f.write(struct.pack("I", length_gltf))
                    f.write("JSON".encode())
                    f.write(gltf_bytes)
                    f.write(b" " * spaces_gltf)

                    # Chunk 1 (BIN)
                    if length_bin > 0:
                        f.write(struct.pack("I", length_bin))
                        f.write("BIN\0".encode())
                        f.write(buffer_bytes)
                        f.write(b"\0" * zeros_bin)

            case False:
                self.export_data(
                    buffer_bytes,
                    filepath=f"{self.export_dir}/{uri_to_path(buffer.uri)}",
                )
                self.export_data(
                    gltf_bytes,
                    filepath=self.export_filepath,
                )

        return self.export_filepath

    def export_data(self, data: bytes, filepath: str) -> str:
        """Safely writes data to destination in export_dir"""
        path = os.path.join(os.path.dirname(self.export_dir), filepath)
        try:
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "wb") as f:
                f.write(data)
            return path
        except Exception as e:
            raise Exception(f"Couldn't export file: {path}") from e

    def export_accessor(self, gltf_data: Gltf2Data, accessor_idx: int) -> bool:
        """Export accessor to buffer"""
        binary_data = ExpBinaryData.from_numpy(arr=gltf_data.accessors[accessor_idx])
        buffer_view = self.export_buffer.add_and_get_view(
            binary_data
        )  # add to buffer and get buffer view
        self.buffer_views.append(buffer_view)

        # Update accessor references in json
        self.gltf["accessors"][accessor_idx]["byteOffset"] = 0
        self.gltf["accessors"][accessor_idx]["bufferView"] = accessor_idx
        return True

    def export_image(self, gltf_data: Gltf2Data, image_idx: int) -> bool:
        """Export image, either by adding to buffer or writing to file"""
        pyimage = gltf_data.gltf["images"][image_idx]
        uri, buffer_view = pyimage.get("uri", None), pyimage.get("bufferView", None)

        assert not (uri is None and buffer_view is None)

        if uri is not None:
            # write to file
            if self.images_encrypted:
                # .jpeg alters some RGB values on export
                # therefore have to use lossless png filetype
                if not uri.endswith(".png"):
                    uri = os.path.splitext(uri)[0] + ".png"
                    gltf_data.gltf["images"][image_idx]["uri"] = uri

            path = os.path.join(self.export_dir, uri_to_path(uri))
            os.makedirs(os.path.dirname(path), exist_ok=True)
            gltf_data.images[image_idx].save(path, compression_level=0)

        elif buffer_view is not None:
            # add to buffer
            binary_data = ExpBinaryData.from_image(image=gltf_data.images[image_idx])
            buffer_view = self.export_buffer.add_and_get_view(binary_data)
            self.buffer_views.append(buffer_view)

            # update image buffer view references in json
            self.gltf["images"][image_idx]["bufferView"] = (
                len(self.buffer_views) - 1
            )  # TODO: make static
        else:
            raise Exception(
                f"Missing image data: {gltf_data.gltf['images'][image_idx]}"
            )
        return True



import sys
import traceback


def from_int(x):
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_none(x):
    assert x is None
    return x


def from_union(fs, x):
    tracebacks = []
    for f in fs:
        try:
            return f(x)
        except AssertionError:
            _, _, tb = sys.exc_info()
            tracebacks.append(tb)
    for tb in tracebacks:
        traceback.print_tb(tb)  # Fixed format
        tb_info = traceback.extract_tb(tb)
        for tbi in tb_info:
            filename, line, func, text = tbi
            print(
                "ERROR",
                "An error occurred on line {} in statement {}".format(line, text),
            )
    assert False


def from_dict(f, x):
    assert isinstance(x, dict)
    return {k: f(v) for (k, v) in x.items()}


def to_class(c, x):
    assert isinstance(x, c)
    return x.to_dict()


def from_list(f, x):
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_float(x):
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)


def from_str(x):
    assert isinstance(x, str)
    return x


def from_bool(x):
    assert isinstance(x, bool)
    return x


def to_float(x):
    assert isinstance(x, float)
    return x


def extension_to_dict(obj):
    if hasattr(obj, "to_list"):
        obj = obj.to_list()
    if hasattr(obj, "to_dict"):
        obj = obj.to_dict()
    if isinstance(obj, list):
        return [extension_to_dict(x) for x in obj]
    elif isinstance(obj, dict):
        return {k: extension_to_dict(v) for (k, v) in obj.items()}
    return obj


def from_extension(x):
    x = extension_to_dict(x)
    assert isinstance(x, dict)
    return x


def from_extra(x):
    return extension_to_dict(x)


class AccessorSparseIndices:
    """Index array of size `count` that points to those accessor attributes that deviate from
    their initialization value. Indices must strictly increase.

    Indices of those attributes that deviate from their initialization value.
    """

    def __init__(self, buffer_view, byte_offset, component_type, extensions, extras):
        self.buffer_view = buffer_view
        self.byte_offset = byte_offset
        self.component_type = component_type
        self.extensions = extensions
        self.extras = extras

    @staticmethod
    def from_dict(obj):
        assert isinstance(obj, dict)
        buffer_view = from_int(obj.get("bufferView"))
        byte_offset = from_union([from_int, from_none], obj.get("byteOffset"))
        component_type = from_int(obj.get("componentType"))
        extensions = from_union(
            [lambda x: from_dict(lambda x: from_dict(lambda x: x, x), x), from_none],
            obj.get("extensions"),
        )
        extras = obj.get("extras")
        return AccessorSparseIndices(
            buffer_view, byte_offset, component_type, extensions, extras
        )

    def to_dict(self):
        result = {}
        result["bufferView"] = from_int(self.buffer_view)
        result["byteOffset"] = from_union([from_int, from_none], self.byte_offset)
        result["componentType"] = from_int(self.component_type)
        result["extensions"] = from_union(
            [lambda x: from_dict(from_extension, x), from_none], self.extensions
        )
        result["extras"] = from_extra(self.extras)
        return result


class AccessorSparseValues:
    """Array of size `count` times number of components, storing the displaced accessor
    attributes pointed by `indices`. Substituted values must have the same `componentType`
    and number of components as the base accessor.

    Array of size `accessor.sparse.count` times number of components storing the displaced
    accessor attributes pointed by `accessor.sparse.indices`.
    """

    def __init__(self, buffer_view, byte_offset, extensions, extras):
        self.buffer_view = buffer_view
        self.byte_offset = byte_offset
        self.extensions = extensions
        self.extras = extras

    @staticmethod
    def from_dict(obj):
        assert isinstance(obj, dict)
        buffer_view = from_int(obj.get("bufferView"))
        byte_offset = from_union([from_int, from_none], obj.get("byteOffset"))
        extensions = from_union(
            [lambda x: from_dict(lambda x: from_dict(lambda x: x, x), x), from_none],
            obj.get("extensions"),
        )
        extras = obj.get("extras")
        return AccessorSparseValues(buffer_view, byte_offset, extensions, extras)

    def to_dict(self):
        result = {}
        result["bufferView"] = from_int(self.buffer_view)
        result["byteOffset"] = from_union([from_int, from_none], self.byte_offset)
        result["extensions"] = from_union(
            [lambda x: from_dict(from_extension, x), from_none], self.extensions
        )
        result["extras"] = from_extra(self.extras)
        return result


class AccessorSparse:
    """Sparse storage of attributes that deviate from their initialization value."""

    def __init__(self, count, extensions, extras, indices, values):
        self.count = count
        self.extensions = extensions
        self.extras = extras
        self.indices = indices
        self.values = values

    @staticmethod
    def from_dict(obj):
        assert isinstance(obj, dict)
        count = from_int(obj.get("count"))
        extensions = from_union(
            [lambda x: from_dict(lambda x: from_dict(lambda x: x, x), x), from_none],
            obj.get("extensions"),
        )
        extras = obj.get("extras")
        indices = AccessorSparseIndices.from_dict(obj.get("indices"))
        values = AccessorSparseValues.from_dict(obj.get("values"))
        return AccessorSparse(count, extensions, extras, indices, values)

    def to_dict(self):
        result = {}
        result["count"] = from_int(self.count)
        result["extensions"] = from_union(
            [lambda x: from_dict(from_extension, x), from_none], self.extensions
        )
        result["extras"] = from_extra(self.extras)
        result["indices"] = to_class(AccessorSparseIndices, self.indices)
        result["values"] = to_class(AccessorSparseValues, self.values)
        return result


class Accessor:
    """A typed view into a bufferView.  A bufferView contains raw binary data.  An accessor
    provides a typed view into a bufferView or a subset of a bufferView similar to how
    WebGL's `vertexAttribPointer()` defines an attribute in a buffer.
    """

    def __init__(
        self,
        buffer_view,
        byte_offset,
        component_type,
        count,
        extensions,
        extras,
        max,
        min,
        name,
        normalized,
        sparse,
        type,
    ):
        self.buffer_view: int = buffer_view
        self.byte_offset = byte_offset
        self.component_type = component_type
        self.count = count
        self.extensions = extensions
        self.extras = extras
        self.max = max
        self.min = min
        self.name = name
        self.normalized = normalized
        self.sparse = sparse
        self.type = type

    @staticmethod
    def from_dict(obj):
        assert isinstance(obj, dict)
        buffer_view = from_union([from_int, from_none], obj.get("bufferView"))
        byte_offset = from_union([from_int, from_none], obj.get("byteOffset"))
        component_type = from_int(obj.get("componentType"))
        count = from_int(obj.get("count"))
        extensions = from_union(
            [lambda x: from_dict(lambda x: from_dict(lambda x: x, x), x), from_none],
            obj.get("extensions"),
        )
        extras = obj.get("extras")
        max = from_union(
            [lambda x: from_list(from_float, x), from_none], obj.get("max")
        )
        min = from_union(
            [lambda x: from_list(from_float, x), from_none], obj.get("min")
        )
        name = from_union([from_str, from_none], obj.get("name"))
        normalized = from_union([from_bool, from_none], obj.get("normalized"))
        sparse = from_union([AccessorSparse.from_dict, from_none], obj.get("sparse"))
        type = from_str(obj.get("type"))
        return Accessor(
            buffer_view,
            byte_offset,
            component_type,
            count,
            extensions,
            extras,
            max,
            min,
            name,
            normalized,
            sparse,
            type,
        )

    def to_dict(self):
        result = {}
        result["bufferView"] = from_union([from_int, from_none], self.buffer_view)
        result["byteOffset"] = from_union([from_int, from_none], self.byte_offset)
        result["componentType"] = from_int(self.component_type)
        result["count"] = from_int(self.count)
        result["extensions"] = from_union(
            [lambda x: from_dict(from_extension, x), from_none], self.extensions
        )
        result["extras"] = from_extra(self.extras)
        result["max"] = from_union(
            [lambda x: from_list(to_float, x), from_none], self.max
        )
        result["min"] = from_union(
            [lambda x: from_list(to_float, x), from_none], self.min
        )
        result["name"] = from_union([from_str, from_none], self.name)
        result["normalized"] = from_union([from_bool, from_none], self.normalized)
        result["sparse"] = from_union(
            [lambda x: to_class(AccessorSparse, x), from_none], self.sparse
        )
        result["type"] = from_str(self.type)
        return result


class BufferView:
    """A view into a buffer generally representing a subset of the buffer."""

    def __init__(
        self,
        buffer,
        byte_length,
        byte_offset,
        byte_stride,
        extensions,
        extras,
        name,
        target,
    ):
        self.buffer = buffer
        self.byte_length = byte_length
        self.byte_offset = byte_offset
        self.byte_stride = byte_stride
        self.extensions = extensions
        self.extras = extras
        self.name = name
        self.target = target

    @staticmethod
    def from_dict(obj):
        assert isinstance(obj, dict)
        buffer = from_int(obj.get("buffer"))
        byte_length = from_int(obj.get("byteLength"))
        byte_offset = from_union([from_int, from_none], obj.get("byteOffset"))
        byte_stride = from_union([from_int, from_none], obj.get("byteStride"))
        extensions = from_union(
            [lambda x: from_dict(lambda x: from_dict(lambda x: x, x), x), from_none],
            obj.get("extensions"),
        )
        extras = obj.get("extras")
        name = from_union([from_str, from_none], obj.get("name"))
        target = from_union([from_int, from_none], obj.get("target"))
        return BufferView(
            buffer,
            byte_length,
            byte_offset,
            byte_stride,
            extensions,
            extras,
            name,
            target,
        )

    def to_dict(self):
        result = {}
        result["buffer"] = from_int(self.buffer)
        result["byteLength"] = from_int(self.byte_length)
        result["byteOffset"] = from_union([from_int, from_none], self.byte_offset)
        result["byteStride"] = from_union([from_int, from_none], self.byte_stride)
        result["extensions"] = from_union(
            [lambda x: from_dict(from_extension, x), from_none], self.extensions
        )
        result["extras"] = from_extra(self.extras)
        result["name"] = from_union([from_str, from_none], self.name)
        result["target"] = from_union([from_int, from_none], self.target)
        return result


class Buffer:
    """A buffer points to binary geometry, animation, or skins."""

    def __init__(self, byte_length, extensions, extras, name, uri):
        self.byte_length = byte_length
        self.extensions = extensions
        self.extras = extras
        self.name = name
        self.uri = uri

    @staticmethod
    def from_dict(obj):
        assert isinstance(obj, dict)
        byte_length = from_int(obj.get("byteLength"))
        extensions = from_union(
            [lambda x: from_dict(lambda x: from_dict(lambda x: x, x), x), from_none],
            obj.get("extensions"),
        )
        extras = obj.get("extras")
        name = from_union([from_str, from_none], obj.get("name"))
        uri = from_union([from_str, from_none], obj.get("uri"))
        return Buffer(byte_length, extensions, extras, name, uri)

    def to_dict(self):
        result = {}
        result["byteLength"] = from_int(self.byte_length)
        result["extensions"] = from_union(
            [lambda x: from_dict(from_extension, x), from_none], self.extensions
        )
        result["extras"] = from_extra(self.extras)
        result["name"] = from_union([from_str, from_none], self.name)
        result["uri"] = from_union([from_str, from_none], self.uri)
        return result

class Image:
    """Image data used to create a texture. Image can be referenced by URI or `bufferView`
    index. `mimeType` is required in the latter case.
    """

    def __init__(self, buffer_view, extensions, extras, mime_type, name, uri):
        self.buffer_view = buffer_view
        self.extensions = extensions
        self.extras = extras
        self.mime_type = mime_type
        self.name = name
        self.uri = uri

    @staticmethod
    def from_dict(obj):
        assert isinstance(obj, dict)
        buffer_view = from_union([from_int, from_none], obj.get("bufferView"))
        extensions = from_union([lambda x: from_dict(lambda x: from_dict(lambda x: x, x), x), from_none],
                                obj.get("extensions"))
        extras = obj.get("extras")
        mime_type = from_union([from_str, from_none], obj.get("mimeType"))
        name = from_union([from_str, from_none], obj.get("name"))
        uri = from_union([from_str, from_none], obj.get("uri"))
        return Image(buffer_view, extensions, extras, mime_type, name, uri)

    def to_dict(self):
        result = {}
        result["bufferView"] = from_union([from_int, from_none], self.buffer_view)
        result["extensions"] = from_union([lambda x: from_dict(from_extension, x), from_none],
                                          self.extensions)
        result["extras"] = from_extra(self.extras)
        result["mimeType"] = from_union([from_str, from_none], self.mime_type)
        result["name"] = from_union([from_str, from_none], self.name)
        result["uri"] = from_union([from_str, from_none], self.uri)
        return result



from time import time

import numpy as np
from crypto_gltf.data.types import EncryptionResponse
from crypto_gltf.encrypt.deprecit.adaptive_v2.aes_gcm import aes_gcm_encrypt
from crypto_gltf.encrypt.deprecit.adaptive_v2.base import AdaptiveBaseModel
from crypto_gltf.encrypt.deprecit.adaptive_v2.types import (
    AdaptiveCipherParamsV2,
    BlockSelection,
    Key,
    SBlocks,
)
from crypto_gltf.encrypt.deprecit.adaptive_v2.utils import (
    combine_sblocks,
    get_sblocks,
    put_sblocks,
)
from loguru import logger


class AdaptiveEncryptionModel(AdaptiveBaseModel):

    @staticmethod
    def _encrypt(
        data: list[np.ndarray], key: Key, params: AdaptiveCipherParamsV2
    ) -> EncryptionResponse[list[np.ndarray], np.ndarray, Key]:

        tic = time()

        sblocks_list = []
        for arr in data:
            sblocks_list.append(
                get_sblocks(arr=arr, params=params, to_get=BlockSelection.all())
            )

        toc = time() - tic
        combined_sblocks = combine_sblocks(sblocks_list=sblocks_list)

        assert key.k1 and key.k2 and key.k3

        r1 = aes_gcm_encrypt(message=combined_sblocks.s1, key=key.k1)
        r2 = aes_gcm_encrypt(message=combined_sblocks.s2, key=key.k2)
        r3 = aes_gcm_encrypt(message=combined_sblocks.s3, key=key.k3)

        r_blocks = SBlocks(s1=r1.ciphertext, s2=r2.ciphertext, s3=r3.ciphertext)

        tic = time()

        encrypted_data = []
        offset = 0
        for arr in data:
            encrypted_data.append(
                put_sblocks(
                    arr=arr,
                    params=params,
                    sblocks=r_blocks,
                    offset=offset,
                    to_put=BlockSelection.all(),
                )
            )
            offset += arr.size

        if not offset == sum([arr.size for arr in data]):
            raise Exception(f"Entire cipher text has not been used")

        aad_arr = np.column_stack(
            (
                np.frombuffer(r1.aad, dtype=np.uint32),
                np.frombuffer(r2.aad, dtype=np.uint32),
                np.frombuffer(r3.aad, dtype=np.uint32),
            )
        )  # convert to arr so we can embed in file
        logger.info(
            f"Byte retrieval/insertion and reshaping took {time()-tic + toc} seconds"
        )

        return EncryptionResponse[list[np.ndarray], np.ndarray, Key](
            ciphertext=encrypted_data, aad=aad_arr, key=key
        )



from __future__ import annotations

from time import time

import numpy as np
from crypto_gltf.data.asset_file_data_types import Gltf2Data
from crypto_gltf.data.types import AAD_DATA, Composition
from crypto_gltf.io.file.base_file import BaseFile
from crypto_gltf.io.file.gltf2.exp.gltf2_exporter import GlTF2Exporter
from crypto_gltf.io.file.gltf2.imp.gltf2_imp_binary_data import (
    BinaryData as ImpBinaryData,
)
from crypto_gltf.io.file.gltf2.imp.gltf2_importer import GlTF2Importer
from crypto_gltf.io.plaintext.plnm import PlnM
from loguru import logger
from PIL import Image


class GLTFFile(BaseFile):
    data: Gltf2Data

    @classmethod
    def load(cls, import_path: str) -> GLTFFile:
        """Load an glTF/glb file"""

        tic = time()
        gltf_importer = GlTF2Importer.from_filepath(
            filepath=import_path,
        )
        for i in range(len(gltf_importer.accessors)):
            ImpBinaryData.decode_accessor(gltf_importer, i, cache=True)

        for i in range(len(gltf_importer.images)):
            ImpBinaryData.decode_image(gltf_importer, i)
        if gltf_importer.glb_buffer:
            filename_ext = ".glb"
            is_glb = True
        else:
            filename_ext = ".gltf"
            is_glb = False

        data = Gltf2Data(
            gltf=gltf_importer.gltf,
            images=gltf_importer.image_cache,
            accessors=gltf_importer.accessor_cache,
            is_glb=is_glb,
        )

        logger.debug(f"{filename_ext[1:]} import took {time()-tic} seconds")
        return cls(data=data, import_path=import_path, filename_ext=filename_ext)

    def save(self, export_dir: str, images_encrypted: bool = False) -> str:
        """Save to glTF/glb file"""

        tic = time()
        gltf_exporter = GlTF2Exporter(
            export_dir=export_dir,
            filename=self.filename,
            is_glb=self.data.is_glb,
            filename_ext=self.filename_ext,
            images_encrypted=images_encrypted,
        )
        export_filepath = gltf_exporter.export(self.data)

        logger.debug(f"{self.filename_ext[1:]} export took {time()-tic} seconds")

        return export_filepath

    @property
    def plnm(self) -> PlnM:
        """'Get plaintext data from GLTFFile"""
        accessors = [accessor for accessor in self.data.accessors.values()]
        images = [
            np.asarray(image, dtype=np.uint8) for image in self.data.images.values()
        ]
        for image in images:
            assert image.dtype == np.uint8
            assert len(image.shape) < 4

        return PlnM(
            meshes=accessors,
            images=images,
            meshes_dim=len(accessors),
            images_dim=len(images),
        )

    def insert_plnm(self, plnm: PlnM) -> None:
        """Insert plaintext in GLTFFile
        assumes the plaintext data came from the same GLTFFile"""

        self.data.accessors = {idx: mesh for idx, mesh in enumerate(plnm.meshes)}
        self.data.images = {
            idx: Image.fromarray(image) for idx, image in enumerate(plnm.images)
        }

    def embed_aad(self, aad: AAD_DATA) -> None:
        """Embed aad data in file for retrieval during decryption"""

        num_accessors = len(self.data.accessors)
        assert Composition.AAD == aad.aad
        self.data.accessors[num_accessors] = aad.aad
        self.data.gltf["accessors"].append(
            {
                "bufferView": num_accessors,
                "byteOffset": 0,
                "componentType": 5125,
                "count": len(aad.aad),
                "type": "VEC3",
            }
        )
        encryption_info = {
            "encrypted": True,
            "accessor": num_accessors,
            "images_encrypted": aad.encrypt_images,
            "meshes_params": aad.meshes_params,
            "images_params": aad.images_params,
        }

        extras = self.data.gltf["asset"].get("extras")
        if extras is not None:
            self.data.gltf["asset"]["extras"]["encryption_info"] = encryption_info
        else:
            self.data.gltf["asset"]["extras"] = {"encryption_info": encryption_info}

    @property
    def aad(self) -> AAD_DATA:
        """Retrieve embedded aad data"""

        encryption_info = self.data.gltf["asset"].get("extras").get("encryption_info")
        if not encryption_info:
            raise Exception(f"No embedded aad data exists, unable to decrypt asset")
        if not encryption_info["encrypted"]:
            raise Exception("Asset not encrypted")

        aad_accessor_idx = encryption_info["accessor"]
        aad = self.data.accessors[aad_accessor_idx]

        # remove embedded aad
        self.data.gltf["accessors"].pop(aad_accessor_idx)
        self.data.accessors.pop(aad_accessor_idx)

        return AAD_DATA(
            aad=aad,
            encrypt_images=encryption_info["images_encrypted"],
            meshes_params=encryption_info["meshes_params"],
            images_params=encryption_info["images_params"],
        )



import os
from time import time
from typing import Literal

import numpy as np
from crypto_gltf.data.types import EncryptionResponse
from crypto_gltf.encrypt.deprecit.adaptive_v1.aes_gcm import aes_gcm_encrypt
from crypto_gltf.encrypt.deprecit.adaptive_v1.base import AdaptiveBaseModel
from crypto_gltf.encrypt.deprecit.adaptive_v1.types import Key
from crypto_gltf.encrypt.deprecit.adaptive_v1.utils import get_bytes, insert_bytes
from loguru import logger


class AdaptiveEncryptionModel(AdaptiveBaseModel):

    @staticmethod
    def _encrypt(
        data: list[np.ndarray], key: Key, position: Literal[0, 1, 2, 3]
    ) -> EncryptionResponse[list[np.ndarray], np.ndarray, Key]:

        tic = time()
        shapes = [arr.shape for arr in data]
        assert key.k1 is not None

        buffer = bytearray(b"")
        for arr in data:
            buffer.extend(get_bytes(arr, position))

        toc = time() - tic

        response = aes_gcm_encrypt(message=buffer, key=key.k1)

        tic = time()

        encrypted_data = []
        offset = 0
        for arr, shape in zip(data, shapes):
            encrypted_data.append(
                insert_bytes(
                    arr=arr,
                    position=position,
                    buffer=response.ciphertext[offset : offset + arr.size],
                ).reshape(shape)
            )
            offset += arr.size

        if not offset == sum([arr.size for arr in data]):
            raise Exception(f"Entire cipher text has not been used")

        single_aad_arr = np.frombuffer(response.aad, dtype=np.uint32)
        aad_arr = np.column_stack(
            (
                np.frombuffer(response.aad, dtype=np.uint32),
                np.zeros((len(single_aad_arr)), dtype=np.uint32),
                np.zeros((len(single_aad_arr)), dtype=np.uint32),
            )
        )  # convert to arr so we can embed in file

        logger.info(
            f"Byte retrieval/insertion and reshaping took {time()-tic + toc} seconds"
        )

        return EncryptionResponse[list[np.ndarray], np.ndarray, Key](
            ciphertext=encrypted_data, aad=aad_arr, key=key
        )



from time import time

import numpy as np
from crypto_gltf.encrypt.deprecit.adaptive_v2.aes_gcm import aes_gcm_decrypt
from crypto_gltf.encrypt.deprecit.adaptive_v2.base import AdaptiveBaseModel
from crypto_gltf.encrypt.deprecit.adaptive_v2.types import (
    AdaptiveCipherParamsV2,
    BlockSelection,
    Key,
    SBlocks,
)
from crypto_gltf.encrypt.deprecit.adaptive_v2.utils import (
    combine_sblocks,
    get_sblocks,
    put_sblocks,
)
from loguru import logger


class AdaptiveDecryptionModel(AdaptiveBaseModel):

    @staticmethod
    def _decrypt(
        data: list[np.ndarray],
        key: Key,
        params: AdaptiveCipherParamsV2,
        aad: np.ndarray,
        selection: BlockSelection,
    ) -> list[np.ndarray]:

        tic = time()

        sblocks_list = []
        for arr in data:
            sblocks_list.append(get_sblocks(arr=arr, params=params, to_get=selection))

        toc = time() - tic

        combined_sblocks = combine_sblocks(sblocks_list=sblocks_list)

        s1, s2, s3 = bytearray(b""), bytearray(b""), bytearray(b"")

        if selection.p:
            assert key.k1  # check key exists
            s1 = aes_gcm_decrypt(
                ciphertext=combined_sblocks.s1, aad_b64=aad[:, 0].tobytes(), key=key.k1
            )
        if selection.q:
            assert key.k2  # check key exists
            s2 = aes_gcm_decrypt(
                ciphertext=combined_sblocks.s2, aad_b64=aad[:, 1].tobytes(), key=key.k2
            )
        if selection.r:
            assert key.k3  # check key exists
            s3 = aes_gcm_decrypt(
                ciphertext=combined_sblocks.s3, aad_b64=aad[:, 2].tobytes(), key=key.k3
            )

        s_blocks = SBlocks(s1=s1, s2=s2, s3=s3)

        tic = time()

        decrypted_data = []
        offset = 0
        for arr in data:
            decrypted_data.append(
                put_sblocks(
                    arr=arr,
                    params=params,
                    sblocks=s_blocks,
                    offset=offset,
                    to_put=selection,
                )
            )
            offset += arr.size

        if not offset == sum([arr.size for arr in data]):
            raise Exception(f"Entire plain text has not been used")

        logger.info(
            f"Byte retrieval/insertion and reshaping took {time()-tic + toc} seconds"
        )

        return decrypted_data



from ctypes import CDLL, c_size_t, c_uint32
from pathlib import Path

import numpy as np


def buffer_length(start: int, stop: int, rows: int, cols: int, slices: int = 1) -> int:
    """Returns length of uint32 buffer required for 2d or 3d mat"""
    slice_length = stop - start
    num_bits = slice_length * rows * cols * slices
    return (num_bits - 1) // 32 + 1


CLIB_FILEPATH = str(Path(__file__).parent.parent.parent.resolve() / "clib/clib.so")

try:
    clib = CDLL(CLIB_FILEPATH)
except:
    raise Exception(f"Error: clib.so library has not been built.")

NP_FLOAT32_ARR_2D = np.ctypeslib.ndpointer(dtype=np.float32, ndim=2, flags="C")
NP_UINT32_ARR_1D = np.ctypeslib.ndpointer(dtype=np.uint32, ndim=1, flags="C")
NP_INT8_ARR_3D = np.ctypeslib.ndpointer(dtype=np.uint8, ndim=3, flags="C")
NP_INT8_ARR_2D = np.ctypeslib.ndpointer(dtype=np.uint8, ndim=2, flags="C")


clib.get_bits32.argtypes = [
    NP_FLOAT32_ARR_2D,
    NP_UINT32_ARR_1D,
    c_size_t,
    c_size_t,
    c_size_t,
    c_size_t,
    c_uint32,
]

clib.put_bits32.argtypes = [
    NP_FLOAT32_ARR_2D,
    NP_UINT32_ARR_1D,
    c_size_t,
    c_size_t,
    c_size_t,
    c_size_t,
]


def get_bits(
    arr: np.ndarray,
    buffer: np.ndarray,
    start: int,
    stop: int,
    fill: int = 0,  # buffer padding value
) -> None:
    """get_bits clib wrapper"""
    assert len(buffer.shape) == 1

    match arr.dtype.itemsize:
        case 4:
            """clib get_bits32 function wrapper"""
            if len(arr.shape) != 2:
                raise Exception(f"get_bits32 not supported for shape {arr.shape}")
            rows, cols = arr.shape[0], arr.shape[1]
            if arr.dtype.itemsize != 4:
                raise Exception(
                    f"dtype {arr.dtype} not supported for 2D get_bits, type must be 32 bits long"
                )
            try:
                clib.get_bits32(
                    arr, buffer, rows, cols, start, stop, fill
                )  # call clib function
            except Exception as e:
                raise RuntimeError("Error running clib.get_bits.") from e
        case 1:
            """clib get_bits8 function wrapper"""
            match len(arr.shape):
                case 3:
                    rows, cols, slices = arr.shape[0], arr.shape[1], arr.shape[2]
                    clib.get_bits8.argtypes = [
                        NP_INT8_ARR_3D,
                        NP_UINT32_ARR_1D,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_uint32,
                    ]
                case 2:
                    rows, cols, slices = arr.shape[0], arr.shape[1], 1
                    clib.get_bits8.argtypes = [
                        NP_INT8_ARR_2D,
                        NP_UINT32_ARR_1D,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_uint32,
                    ]
                case _:
                    raise Exception(f"get_bits8 not supported for shape {arr.shape}")
            try:
                clib.get_bits8(
                    arr, buffer, rows, cols, slices, start, stop, fill
                )  # call clib function
            except Exception as e:
                raise RuntimeError("Error running clib.get_bits.") from e
        case _:
            raise Exception(
                f"get_bits not supported for dtype {arr.dtype.itemsize}, bit length must be 8 or 32."
            )


def put_bits(
    putarr: np.ndarray,
    buffer: np.ndarray,
    start: int,
    stop: int,
) -> None:
    """put_bits clib wrapper"""
    assert len(buffer.shape) == 1

    match putarr.dtype.itemsize:
        case 4:
            """clib put_bits32 function wrapper"""
            if len(putarr.shape) != 2:
                raise Exception(f"put_bits32 not supported for shape {putarr.shape}")
            rows, cols = putarr.shape[0], putarr.shape[1]
            if putarr.dtype.itemsize != 4:
                raise Exception(
                    f"dtype {putarr.dtype} not supported for 2D put_bits, type must be 32 bits long"
                )
            try:
                clib.put_bits32(
                    putarr, buffer, rows, cols, start, stop
                )  # call clib function
            except Exception as e:
                raise RuntimeError("Error running clib.put_bits.") from e
        case 1:
            """clib put_bits8 function wrapper"""
            match len(putarr.shape):
                case 3:
                    rows, cols, slices = (
                        putarr.shape[0],
                        putarr.shape[1],
                        putarr.shape[2],
                    )
                    clib.put_bits8.argtypes = [
                        NP_INT8_ARR_3D,
                        NP_UINT32_ARR_1D,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                    ]
                case 2:
                    rows, cols, slices = putarr.shape[0], putarr.shape[1], 1
                    clib.put_bits8.argtypes = [
                        NP_INT8_ARR_2D,
                        NP_UINT32_ARR_1D,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                        c_size_t,
                    ]
                case _:
                    raise Exception(f"put_bits8 not supported for shape {putarr.shape}")
            try:
                clib.put_bits8(
                    putarr, buffer, rows, cols, slices, start, stop
                )  # call clib function
            except Exception as e:
                raise RuntimeError("Error running clib.put_bits.") from e
        case _:
            raise Exception(
                f"put_bits not supported for dtype {putarr.dtype.itemsize}, bit length must be 8 or 32."
            )



import secrets

import numpy as np
from crypto_gltf.encrypt.adaptive.types import (
    AdaptiveCipherParams,
    BlockSelection,
    BufferView,
    Key,
)
from crypto_gltf.encrypt.adaptive.utils import get_bits
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from loguru import logger


def aes_sha(sblock: bytes, ki: bytes) -> bytes:
    """Generate subsequent subkey k_(i-1) using AES followed by sha"""

    assert len(sblock) == 32

    cipher = Cipher(algorithms.AES(ki), modes.ECB(), default_backend())
    encryptor = cipher.encryptor()
    ciphertext = encryptor.update(sblock) + encryptor.finalize()

    digest = hashes.Hash(hashes.SHA256())
    digest.update(sblock)
    digest.update(ciphertext)

    return digest.finalize()


def get_source_arrays(plnm_arrs: list[np.ndarray], min_size: int) -> list[np.ndarray]:
    """Return first n matrices in plnm_arrs such that the sum of matrix sizes is greater than min_size.
    256 = AES keylength"""

    source_arrays: list[np.ndarray] = []

    for arr in plnm_arrs:
        source_arrays.append(arr)
        size = sum([arr.size for arr in source_arrays])
        if size > min_size:
            break

    if sum([arr.size for arr in source_arrays]) < min_size:
        raise Exception("Mesh too small to generate keys")

    return source_arrays


def get_min_size_plaintext(
    params: AdaptiveCipherParams, selection: BlockSelection
) -> int:
    """Returns minimum number of plaintext float32 values required to generate 256 bit subkeys"""
    smallest_selected_param = min(params.selection(selection).values())
    return 256 // smallest_selected_param + 1


def generate_keys(
    plnm_arrs: list[np.ndarray], params: AdaptiveCipherParams, k3: bytes | None = None
) -> Key:
    """Calculates k1, k2 and k3 for a given plaintext asset"""

    min_size = get_min_size_plaintext(
        params=params, selection=BlockSelection(q=True, r=True)
    )
    source_arrays = get_source_arrays(
        plnm_arrs=plnm_arrs,
        min_size=min_size,
    )

    s2, s3 = bytearray(b""), bytearray(b"")

    for arr in source_arrays:

        buffer_view = BufferView.from_shape(shape=arr.shape, params=params)

        q_buffer, r_buffer = (
            buffer_view.q_buffer,
            buffer_view.r_buffer,
        )
        get_bits(arr, q_buffer, params.qstart, params.qstop)
        get_bits(arr, r_buffer, params.rstart, params.rstop)

        s2.extend(q_buffer.tobytes())
        s3.extend(r_buffer.tobytes())

    s3 = s3[:32]
    s2 = s2[:32]

    if k3:
        assert len(k3) == 32
        logger.info("User specified key detected, using key provided.")
    else:
        k3 = secrets.token_bytes(32)

    k2 = aes_sha(sblock=s3, ki=k3)
    k1 = aes_sha(sblock=s2, ki=k2)

    return Key(k1=k1, k2=k2, k3=k3)


def get_subkey(
    plnm_arrs: list[np.ndarray],
    params: AdaptiveCipherParams,
    ki: bytes,
    selection: BlockSelection,
) -> bytes:
    """Calculate ki-1 from ki for a given plaintext"""

    min_size = get_min_size_plaintext(params, selection)
    source_arrays = get_source_arrays(plnm_arrs=plnm_arrs, min_size=min_size)

    block = selection.single_block
    start = params.start(block)
    stop = params.stop(block)

    si = bytearray(b"")
    for arr in source_arrays:
        buffer_view = BufferView.from_shape(arr.shape, params=params)
        buffer = buffer_view.buffer(block)
        get_bits(arr, buffer, start, stop)
        si.extend(buffer.tobytes())

    ki_minus_1 = aes_sha(sblock=si[:32], ki=ki)

    return ki_minus_1



from typing import Literal

from crypto_gltf.data.types import BaseKey, BaseParams


class AdaptiveCipherParamsV1(BaseParams):
    p: Literal[0, 1, 2, 3]
    q: int = 0
    r: int = 0


class Key(BaseKey): ...



from time import time
from typing import Literal

import numpy as np
from crypto_gltf.encrypt.deprecit.adaptive_v1.aes_gcm import aes_gcm_decrypt
from crypto_gltf.encrypt.deprecit.adaptive_v1.base import AdaptiveBaseModel
from crypto_gltf.encrypt.deprecit.adaptive_v1.types import Key
from crypto_gltf.encrypt.deprecit.adaptive_v1.utils import get_bytes, insert_bytes
from loguru import logger


class AdaptiveDecryptionModel(AdaptiveBaseModel):

    @staticmethod
    def _decrypt(
        data: list[np.ndarray],
        key: Key,
        position: Literal[0, 1, 2, 3],
        aad: np.ndarray,
    ) -> list[np.ndarray]:
        tic = time()

        shapes = [arr.shape for arr in data]

        buffer = bytearray(b"")
        for arr in data:
            buffer.extend(get_bytes(arr, position))

        toc = time() - tic

        assert key.k1
        decrypted_buffer = aes_gcm_decrypt(
            ciphertext=buffer, aad_b64=aad.tobytes(), key=key.k1
        )

        tic = time()

        decrypted_data = []
        offset = 0
        for arr, shape in zip(data, shapes):
            decrypted_data.append(
                insert_bytes(
                    arr=arr,
                    position=position,
                    buffer=decrypted_buffer[offset : offset + arr.size],
                ).reshape(shape)
            )
            offset += arr.size

        logger.info(
            f"Byte retrieval/insertion and reshaping took {time()-tic + toc} seconds"
        )

        return decrypted_data



from __future__ import annotations

from base64 import urlsafe_b64encode as b64e
from functools import cached_property
from typing import Any, Literal

import numpy as np
from crypto_gltf.data.types import BaseKey, BaseParams
from crypto_gltf.encrypt.adaptive.utils import buffer_length
from pydantic import BaseModel, model_validator


class Key(BaseKey):

    @model_validator(mode="before")
    @classmethod
    def validate_fields(cls, data: Any):
        k1, k2, k3 = data.get("k1"), data.get("k2"), data.get("k3")
        assert k1 or k2 or k3  # at least one key must be provided
        return data

    def __str__(self) -> str:
        return str({k: b64e(v) for k, v in self.__dict__.items() if v is not None})

    @property
    def filled(self) -> bool:
        """Returns true if all subkeys below highest non-null subkey exist"""
        if self.k3 is not None:
            return (self.k2 is not None) and (self.k1 is not None)
        elif self.k2 is not None:
            return self.k1 is not None
        else:
            return self.k1 is not None  # could remove but kept for clarity

    @property
    def size(self) -> int:
        """Returns the number of non null subkeys"""
        return len([ki for ki in self.__dict__.values() if ki is not None])


class BlockSelection(BaseModel):
    p: bool = False
    q: bool = False
    r: bool = False

    @classmethod
    def all(cls) -> BlockSelection:
        """Select all blocks"""
        return cls(p=True, q=True, r=True)

    @property
    def single_block(self) -> Literal["p", "q", "r"]:
        assert sum([self.p, self.q, self.r]) == 1
        if self.p:
            return "p"
        elif self.q:
            return "q"
        else:
            return "r"


class AdaptiveCipherParams(BaseParams):
    p: int
    q: int
    r: int

    @cached_property
    def pstart(self) -> int:
        raise NotImplementedError()

    @cached_property
    def pstop(self) -> int:
        raise NotImplementedError()

    @cached_property
    def qstart(self) -> int:
        raise NotImplementedError()

    @cached_property
    def qstop(self) -> int:
        raise NotImplementedError()

    @cached_property
    def rstart(self) -> int:
        raise NotImplementedError()

    @cached_property
    def rstop(self) -> int:
        raise NotImplementedError()

    def start(self, block: Literal["p", "q", "r"]) -> int:
        MAPPING = {"p": "pstart", "q": "qstart", "r": "rstart"}
        return self.__getattribute__(MAPPING[block])

    def stop(self, block: Literal["p", "q", "r"]) -> int:
        MAPPING = {"p": "pstop", "q": "qstop", "r": "rstop"}
        return self.__getattribute__(MAPPING[block])

    def selection(self, selection: BlockSelection) -> dict:
        """'Returns selected params"""
        return {k: self.__getattribute__(k) for k, v in selection.__dict__.items() if v}


class ImagesAdaptiveCipherParams(AdaptiveCipherParams):

    @model_validator(mode="before")
    @classmethod
    def validate_fields(cls, data: Any):
        p, q, r = data.get("p"), data.get("q"), data.get("r")
        assert p and q and r  # all fields must be provided
        assert p >= 1
        assert q >= 1
        assert r >= 1
        assert p + q + r == 8
        return data

    @cached_property
    def pstart(self) -> int:
        return self.qstart - self.p

    @cached_property
    def pstop(self) -> int:
        return self.qstart

    @cached_property
    def qstart(self) -> int:
        return self.qstop - self.q

    @cached_property
    def qstop(self) -> int:
        return self.rstart

    @cached_property
    def rstart(self) -> int:
        return 8 - self.r

    @cached_property
    def rstop(self) -> int:
        return 8


class MeshesAdaptiveCipherParams(AdaptiveCipherParams):

    @model_validator(mode="before")
    @classmethod
    def validate_fields(cls, data: Any):
        p, q, r = data.get("p"), data.get("q"), data.get("r")
        assert p and q and r  # all fields must be provided
        assert p >= 1
        assert q >= 1
        assert r >= 1
        assert p + q + r <= 23
        return data

    @cached_property
    def pstart(self) -> int:
        return 9

    @cached_property
    def pstop(self) -> int:
        return 9 + self.p

    @cached_property
    def qstart(self) -> int:
        return self.pstop

    @cached_property
    def qstop(self) -> int:
        return self.qstart + self.q

    @cached_property
    def rstart(self) -> int:
        return self.qstop

    @cached_property
    def rstop(self) -> int:
        return self.rstart + self.r


class BufferView(BaseModel):

    params: AdaptiveCipherParams
    rows: int
    cols: int
    slices: int = 1

    @classmethod
    def from_shape(
        cls, shape: tuple[int, int] | tuple[int, int, int], params: AdaptiveCipherParams
    ) -> BufferView:
        if len(shape) == 2:
            return cls(rows=shape[0], cols=shape[1], params=params)
        elif len(shape) == 3:
            return cls(rows=shape[0], cols=shape[1], slices=shape[2], params=params)

    @cached_property
    def pbufflen(self) -> int:
        """p uint32 buffer length required"""
        return buffer_length(
            self.params.pstart, self.params.pstop, self.rows, self.cols, self.slices
        )

    @cached_property
    def qbufflen(self) -> int:
        """q uint32 buffer length required"""
        return buffer_length(
            self.params.qstart, self.params.qstop, self.rows, self.cols, self.slices
        )

    @cached_property
    def rbufflen(self) -> int:
        """r uint32 buffer length required"""
        return buffer_length(
            self.params.rstart, self.params.rstop, self.rows, self.cols, self.slices
        )

    @property
    def p_buffer(self) -> np.ndarray:
        return np.zeros((self.pbufflen), dtype=np.uint32, order="C")

    @property
    def q_buffer(self) -> np.ndarray:
        return np.zeros((self.qbufflen), dtype=np.uint32, order="C")

    @property
    def r_buffer(self) -> np.ndarray:
        return np.zeros((self.rbufflen), dtype=np.uint32, order="C")

    def buffer(self, block: Literal["p", "q", "r"]) -> np.ndarray:
        MAPPING = {"p": "p_buffer", "q": "q_buffer", "r": "r_buffer"}
        return self.__getattribute__(MAPPING[block])

    def bufflen(self, block: Literal["p", "q", "r"]) -> np.ndarray:
        MAPPING = {"p": "pbufflen", "q": "qbufflen", "r": "rbufflen"}
        return self.__getattribute__(MAPPING[block])



from __future__ import annotations

import numpy as np
from pydantic import BaseModel, ConfigDict


class CACipherParams(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    gens: int
    ps: list[int]
    idl_points: np.ndarray = np.empty(0, dtype=int)

    def validate_idl_points(self, dim: int) -> None:
        # use default idl points if undefined
        if len(self.idl_points) == 0:
            self.idl_points = np.zeros((dim, 2), dtype=int)
        if len(self.idl_points) != dim:
            raise Exception(f"Number of plaintext matrices != number of idl points")


class CABaseModel:
    """Base model for Conways game of life cellular automata encryption/decryption model"""

    def __new__(cls, *args, **kwargs):
        raise RuntimeError(f"{cls} should not be instantiated")

    @staticmethod
    def get_nonzero_idxs(
        expCA: np.ndarray, mat: np.ndarray, idl: tuple[int, int]
    ) -> np.ndarray:
        """Get flattened indices of non-zero expCA cells"""

        row_idxs = np.arange(idl[0], idl[0] + mat.shape[0]) % expCA.shape[0]
        col_idxs = np.arange(idl[1], idl[1] + mat.shape[1]) % expCA.shape[1]

        covered_expCA = expCA[np.ix_(row_idxs, col_idxs)]
        flat_covered_expCA = covered_expCA.flatten()
        nonzero_elts_flat_idxs = np.nonzero(flat_covered_expCA)[0]

        return nonzero_elts_flat_idxs

    @staticmethod
    def get_zero_idxs(
        nonzero_elts_flat_idxs: np.ndarray, mat: np.ndarray
    ) -> np.ndarray:
        """Get flattened indices of zero expCA cells"""

        zero_elts_flat_idxs = np.setdiff1d(np.arange(mat.size), nonzero_elts_flat_idxs)
        if len(zero_elts_flat_idxs) + len(nonzero_elts_flat_idxs) != mat.size:
            raise Exception(
                f"Mismatched confused/resolved matrix and covered expCA ndarray sizes: \
                        {len(zero_elts_flat_idxs) + len(nonzero_elts_flat_idxs)} != {mat.size}"
            )
        return zero_elts_flat_idxs



from pathlib import Path
from typing import Type

from crypto_gltf.io.file.base_file import BaseFile
from crypto_gltf.io.file.gltf2.gltf2 import GLTFFile
from crypto_gltf.io.file.off.off import OffFile


def File(import_filepath: str) -> BaseFile:
    """Factory method for initializing any filetype"""

    localizers: dict[str, Type[BaseFile]] = {
        ".gltf": GLTFFile,
        ".glb": GLTFFile,
        ".off": OffFile,
    }
    path = Path(import_filepath)

    filetype = localizers.get(path.suffix.lower())
    if not filetype:
        raise Exception(f"Filetype {filetype} not supported.")

    return filetype.load(import_filepath)



from __future__ import annotations

from time import time

import numpy as np
from crypto_gltf.data.types import AAD_DATA, EncryptionResponse
from crypto_gltf.encrypt.deprecit.adaptive_v2.decrypt import AdaptiveDecryptionModel
from crypto_gltf.encrypt.deprecit.adaptive_v2.encrypt import AdaptiveEncryptionModel
from crypto_gltf.encrypt.deprecit.adaptive_v2.key_gen import (
    generate_keys,
    get_k1,
    get_k2,
)
from crypto_gltf.encrypt.deprecit.adaptive_v2.types import (
    AdaptiveCipherParamsV2,
    BlockSelection,
    Key,
)
from crypto_gltf.io.plaintext.plnm import PlnM
from loguru import logger


class AdaptiveCryptoSystemV2(AdaptiveEncryptionModel, AdaptiveDecryptionModel):
    """Adaptive cipher system"""

    AAD: bool = True

    @staticmethod
    def encrypt(
        plnm: PlnM,
        meshes_cipher_params: AdaptiveCipherParamsV2,
        images_cipher_params: AdaptiveCipherParamsV2,
        encrypt_images: bool = False,
        key: Key | None = None,
    ) -> EncryptionResponse[PlnM, AAD_DATA, Key]:

        tic = time()
        if not key:
            key = generate_keys(plnm, meshes_cipher_params)
        else:
            assert key.size == 1 and key.k3 is not None
            key = generate_keys(plnm, meshes_cipher_params, k3=key.k3)
        # Get non-integer type arrays
        # i.e. vertices and colors
        meshes_float_arrs_idxs = np.array(
            [
                idx
                for idx, mesh in enumerate(plnm.meshes)
                if mesh.dtype.kind == "f" and mesh.size > 0
            ]
        )
        float_arrs = [
            plnm.meshes[i].astype(np.float32) for i in meshes_float_arrs_idxs
        ]  # convert to float32

        encryption_response = AdaptiveEncryptionModel._encrypt(
            data=float_arrs, key=key, params=meshes_cipher_params
        )

        for idx, i in enumerate(meshes_float_arrs_idxs):
            plnm.meshes[i] = encryption_response.ciphertext[idx]

        aad = AAD_DATA(
            aad=encryption_response.aad,
            encrypt_images=encrypt_images,
            meshes_params=meshes_cipher_params.__dict__,
            images_params=images_cipher_params.__dict__,
        )

        logger.debug(f"Adaptive encryption took {time()-tic} seconds.")
        return EncryptionResponse[PlnM, AAD_DATA, Key](
            ciphertext=plnm, aad=aad, key=key
        )

    @staticmethod
    def decrypt(
        plnm: PlnM,
        key: Key,
        aad: AAD_DATA,
    ) -> PlnM:
        tic = time()
        meshes_cipher_params = AdaptiveCipherParamsV2(**aad.meshes_params)

        if key.size == 0:
            raise Exception(f"Please provide a key")

        # Get non-integer type arrays
        # i.e. vertices and colors
        meshes_float_arrs_idxs = np.array(
            [
                idx
                for idx, mesh in enumerate(plnm.meshes)
                if mesh.dtype.kind == "f" and mesh.size > 0
            ]
        )

        float_arrs = [plnm.meshes[i] for i in meshes_float_arrs_idxs]

        if key.k3:
            float_arrs = AdaptiveDecryptionModel._decrypt(
                data=float_arrs,
                key=key,
                params=meshes_cipher_params,
                aad=aad.aad,
                selection=BlockSelection(r=True),
            )

            toc = time()
            key.k2 = get_k2(
                float_arrs=float_arrs, params=meshes_cipher_params, k3=key.k3
            )

        if key.k2:
            float_arrs = AdaptiveDecryptionModel._decrypt(
                data=float_arrs,
                key=key,
                params=meshes_cipher_params,
                aad=aad.aad,
                selection=BlockSelection(q=True),
            )

            key.k1 = get_k1(
                float_arrs=float_arrs, params=meshes_cipher_params, k2=key.k2
            )

        assert key.k1 is not None

        decrypted_data = AdaptiveDecryptionModel._decrypt(
            data=float_arrs,
            key=key,
            params=meshes_cipher_params,
            aad=aad.aad,
            selection=BlockSelection(p=True),
        )

        for idx, i in enumerate(meshes_float_arrs_idxs):
            plnm.meshes[i] = decrypted_data[idx]

        assert key.filled

        logger.debug(f"Adaptive decryption took {time()-tic} seconds.")

        return plnm



from __future__ import annotations

import base64
import json
import os
import struct
from dataclasses import dataclass

import numpy as np
from crypto_gltf.data.types import JSONDict
from crypto_gltf.io.file.gltf2.com.gltf2_io import Accessor, Buffer, BufferView, Image
from crypto_gltf.io.file.gltf2.utils import uri_to_path
from PIL import Image as PILImage
from pydantic import BaseModel, ConfigDict


class GlbModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    gltf: JSONDict
    glb_buffer: memoryview | None


class DataChunkModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    data_type: bytes
    data_length: int
    data: memoryview
    offset: int


@dataclass
class GlTF2Importer:
    filepath: str
    gltf: JSONDict

    accessors: list[Accessor]
    buffers: dict[int, Buffer]
    buffer_views: dict[int, BufferView]
    images: dict[int, Image]

    glb_buffer: memoryview | None
    loaded_buffers: dict[int, memoryview]

    accessor_cache: dict[int, np.ndarray]
    image_cache: dict[int, PILImage.Image]

    @classmethod
    def from_filepath(cls, filepath: str) -> GlTF2Importer:
        """Load glTF from filepath"""
        if not os.path.isfile(filepath):
            raise ImportError("Please select a file")

        with open(filepath, "rb") as f:
            content = memoryview(f.read())

        if content[:4] == b"glTF":
            glb_model = cls.load_glb(content)
            gltf, glb_buffer = glb_model.gltf, glb_model.glb_buffer
        else:
            gltf = cls.load_json(content)
            glb_buffer = None

        cls.check_version(gltf)

        if gltf.get("accessors"):
            accessors = [Accessor.from_dict(accessor) for accessor in gltf["accessors"]]
        else:
            accessors = []
        if gltf.get("buffers"):
            buffers = [Buffer.from_dict(buffer) for buffer in gltf["buffers"]]
        else:
            buffers = []
        if gltf.get("bufferViews"):
            buffer_views = [
                BufferView.from_dict(bufferview) for bufferview in gltf["bufferViews"]
            ]
        else:
            buffer_views = []
        if "images" in gltf.keys():
            images = [Image.from_dict(image) for image in gltf["images"]]
        else:
            images = []

        return cls(
            filepath=filepath,
            gltf=gltf,
            accessors=accessors,
            buffers={idx: buffer for idx, buffer in enumerate(buffers)},
            buffer_views={
                idx: buffer_view for idx, buffer_view in enumerate(buffer_views)
            },
            images={idx: image for idx, image in enumerate(images)},
            glb_buffer=glb_buffer,
            accessor_cache={},
            loaded_buffers={},
            image_cache={},
        )

    def load_buffer(self, buffer_idx: int) -> None:
        """Load buffer."""
        buffer = self.buffers[buffer_idx]

        if buffer.uri:
            data = self.load_uri(buffer.uri)
            if data is None:
                raise ImportError("Missing resource, '" + buffer.uri + "'.")
            self.loaded_buffers[buffer_idx] = data

        else:
            # GLB-stored buffer
            if buffer_idx == 0 and self.glb_buffer is not None:
                self.loaded_buffers[buffer_idx] = self.glb_buffer

    def load_uri(self, uri: str) -> memoryview:
        """Loads a URI."""
        sep = ";base64,"
        if uri.startswith("data:"):
            idx = uri.find(sep)
            if idx != -1:
                data = uri[idx + len(sep) :]
                return memoryview(base64.b64decode(data))

        path = os.path.join(os.path.dirname(self.filepath), uri_to_path(uri))
        try:
            with open(path, "rb") as f_:
                return memoryview(f_.read())
        except Exception as e:
            raise Exception(f"Couldn't read file: {path}") from e

    @staticmethod
    def load_glb(content: memoryview) -> GlbModel:
        """Load binary glb."""

        magic = content[:4]
        if magic != b"glTF":
            raise ImportError("This file is not a glTF/glb file")

        version, file_size = struct.unpack_from("<II", content, offset=4)
        if version != 2:
            raise ImportError("GLB version must be 2; got %d" % version)
        if file_size != len(content):
            raise ImportError("Bad GLB: file size doesn't match")

        glb_buffer = None
        offset = 12  # header size = 12

        # JSON chunk is first
        data_chunk = GlTF2Importer.load_chunk(content, offset)
        json_bytes = data_chunk.data

        if data_chunk.data_type != b"JSON":
            raise ImportError("Bad GLB: first chunk not JSON")
        if data_chunk.data_length != len(json_bytes):
            raise ImportError("Bad GLB: length of json chunk doesn't match")
        gltf = GlTF2Importer.load_json(json_bytes)

        # BIN chunk is second (if it exists)
        if data_chunk.offset < len(content):
            data_chunk = GlTF2Importer.load_chunk(content, data_chunk.offset)

            if data_chunk.data_type == b"BIN\0":
                if data_chunk.data_length != len(data_chunk.data):
                    raise ImportError("Bad GLB: length of BIN chunk doesn't match")
                glb_buffer = data_chunk.data

        return GlbModel(gltf=gltf, glb_buffer=glb_buffer)

    @staticmethod
    def load_chunk(content: memoryview, offset: int) -> DataChunkModel:
        """Load chunk."""
        chunk_header = struct.unpack_from("<I4s", content, offset)
        data_length = chunk_header[0]
        data_type = chunk_header[1]
        data = content[offset + 8 : offset + 8 + data_length]

        return DataChunkModel(
            data_type=data_type,
            data_length=data_length,
            data=data,
            offset=offset + 8 + data_length,
        )

    @staticmethod
    def load_json(content: memoryview) -> JSONDict:
        def bad_constant(val):
            raise ImportError("Bad glTF: json contained %s" % val)

        try:
            text = str(content, encoding="utf-8")
            return json.loads(text, parse_constant=bad_constant)
        except ValueError as e:
            raise ImportError("Bad glTF: json error: %s" % e.args[0])

    @staticmethod
    def check_version(gltf: JSONDict) -> None:
        """Check version.."""
        if not isinstance(gltf, dict) or "asset" not in gltf:
            raise ImportError("Bad glTF: no asset in json")
        if "version" not in gltf["asset"]:
            raise ImportError("Bad glTF: no version")
        if gltf["asset"]["version"] != "2.0":
            raise ImportError(
                "glTF version must be 2.0; got %s" % gltf["asset"]["version"]
            )



def bit_slice(num: int, start: int, stop: int, length: int = 32) -> int:
    """Returns bit slice for integer"""
    mask = 2 ** (stop - start) - 1
    shift = length - (stop - start) - start
    return (num & (mask << shift)) >> shift



from typing import Any

from dataclasses import dataclass

@dataclass
class BasePlainText:
    """Base class for all plaintext types"""

    meshes: Any
    images: Any

    meshes_dim: int
    images_dim: int



from __future__ import annotations

import numpy as np
from crypto_gltf.encrypt.deprecit.ca.base import CABaseModel, CACipherParams
from crypto_gltf.encrypt.deprecit.ca.ca import ExpConwaysHCA
from loguru import logger


class CAEncryptionModel(CABaseModel):
    """Conways game of life cellular automata encryption model"""

    @staticmethod
    def compute_gens(
        expHCA: ExpConwaysHCA, cipher_params: CACipherParams
    ) -> ExpConwaysHCA:
        """Encrypt data"""

        for i in range(cipher_params.gens - 1):
            logger.info(f"Calculating generation {i+1}")
            if i in cipher_params.ps:
                expHCA.insert_ps()
            else:
                expHCA.diffuse()

        return expHCA

    @staticmethod
    def encrypt(
        plain_text: list[np.ndarray],
        expHCA: ExpConwaysHCA,
        cipher_params: CACipherParams,
    ) -> list[np.ndarray]:
        """Encrypt data"""
        for i in range(cipher_params.gens):
            if i in cipher_params.ps:
                continue
            else:
                plain_text = CAEncryptionModel.confuse_all_pt_matrices(
                    plain_text=plain_text,
                    expCA=expHCA.expHCA[i],
                    idl_points=cipher_params.idl_points,
                )
        return plain_text

    @staticmethod
    def confuse_all_pt_matrices(
        plain_text: list[np.ndarray],
        expCA: np.ndarray,
        idl_points: np.ndarray,
    ) -> list[np.ndarray]:
        """Confuse all plaintext matrices, conducts a single confuse step"""
        confused_data = [
            CAEncryptionModel.confuse_single_pt_matrix(
                expCA=expCA, plain_text=mat, idl=idl
            )
            for mat, idl in zip(plain_text, idl_points)
        ]
        return confused_data

    @staticmethod
    def confuse_single_pt_matrix(
        expCA: np.ndarray, plain_text: np.ndarray, idl: tuple[int, int]
    ) -> np.ndarray:
        """Confuse a single plain_text matrix. One confusion step.
        Split into 2D matrices and send to confuse_2d_matrix"""

        shape = plain_text.shape
        dims = len(shape)
        match dims:
            case 0:
                return plain_text
            case 1:
                return plain_text
            case 2:
                return CAEncryptionModel.confuse_2d_matrix(
                    expCA=expCA, plain_text=plain_text, idl=idl
                )
            case 3:
                confused_mat = np.empty(shape, dtype=plain_text.dtype)
                for i in range(shape[-1]):
                    confused_mat[:, :, i] = CAEncryptionModel.confuse_2d_matrix(
                        expCA=expCA, plain_text=plain_text[:, :, i], idl=idl
                    )
                return confused_mat
            case _:
                raise Exception(
                    f"Matrix has too many dimensions: {dims}, for CA encryptioon"
                )

    @staticmethod
    def confuse_2d_matrix(
        expCA: np.ndarray, plain_text: np.ndarray, idl: tuple[int, int]
    ) -> np.ndarray:
        """Confuse a single 2D plain_text matrix. One confusion step"""
        if plain_text.size == 0:
            return plain_text
        confused_mat = np.zeros(plain_text.shape, dtype=plain_text.dtype)
        flat_plain_text = plain_text.flatten()

        nonzero_elts_flat_idxs = CAEncryptionModel.get_nonzero_idxs(
            expCA=expCA, mat=plain_text, idl=idl
        )
        zero_elts_flat_idxs = CAEncryptionModel.get_zero_idxs(
            nonzero_elts_flat_idxs=nonzero_elts_flat_idxs, mat=plain_text
        )

        confused_mat.put(
            nonzero_elts_flat_idxs,
            flat_plain_text[: len(nonzero_elts_flat_idxs)],
        )
        confused_mat.put(
            zero_elts_flat_idxs,
            flat_plain_text[len(nonzero_elts_flat_idxs) :],
        )

        return confused_mat



from time import time

import numpy as np
from crypto_gltf.encrypt.adaptive.base import AdaptiveBaseModel
from crypto_gltf.encrypt.adaptive.cryptography.aes_gcm import aes_gcm_decrypt
from crypto_gltf.encrypt.adaptive.types import (
    AdaptiveCipherParams,
    BlockSelection,
    BufferView,
    Key,
)
from crypto_gltf.encrypt.adaptive.utils import get_bits, put_bits
from loguru import logger


class AdaptiveDecryptionModel(AdaptiveBaseModel):

    @staticmethod
    def _decrypt(
        data: list[tuple[np.ndarray, AdaptiveCipherParams]],
        key: Key,
        selection: BlockSelection,
        aad: np.ndarray,
    ) -> list[tuple[np.ndarray, AdaptiveCipherParams]]:
        tic = time()

        block = selection.single_block

        assert len(aad) == len(data) + 14  # +14 for 14 rows of aad data

        match selection.single_block:
            case "p":
                assert key.k1  # check key exists
                subkey = key.k1
                padding_arr = aad[:, 0]
            case "q":
                assert key.k2  # check key exists
                subkey = key.k2
                padding_arr = aad[:, 1]
            case "r":
                assert key.k3  # check key exists
                subkey = key.k3
                padding_arr = aad[:, 2]
            case _:
                raise Exception("No blocks selected.")

        aad_b64 = padding_arr[-14:].tobytes()

        combined_sblocks = bytearray(b"")
        for idx, (arr, params) in enumerate(data):

            start = params.start(block)
            stop = params.stop(block)
            buffer_view = BufferView.from_shape(shape=arr.shape, params=params)
            buffer = buffer_view.buffer(block=block)
            get_bits(arr, buffer, start, stop)
            buffer[-1] = padding_arr[idx]
            combined_sblocks.extend(buffer.tobytes())

        toc = time() - tic

        try:
            decrypted_data = aes_gcm_decrypt(
                ciphertext=combined_sblocks, aad_b64=aad_b64, key=subkey
            )
        except:
            raise Exception("Invalid key.")

        tic = time()

        decrypted_data_arr = np.frombuffer(decrypted_data, dtype=np.uint32)

        offset = 0
        for arr, params in data:
            buffer_view = BufferView.from_shape(shape=arr.shape, params=params)
            start = params.start(block)
            stop = params.stop(block)

            put_bits(
                arr,
                decrypted_data_arr[offset : offset + buffer_view.bufflen(block=block)],
                start,
                stop,
            )
            offset += buffer_view.bufflen(block=block)

        if not offset == len(decrypted_data_arr):
            raise Exception(f"Entire plain text has not been used")

        logger.debug(
            f"Byte retrieval/insertion and reshaping took {time()-tic + toc} seconds"
        )
        return data



from __future__ import annotations

import secrets
from time import time

import numpy as np
from crypto_gltf.data.types import AAD_DATA, EncryptionResponse
from crypto_gltf.encrypt.deprecit.adaptive_v1.decrypt import AdaptiveDecryptionModel
from crypto_gltf.encrypt.deprecit.adaptive_v1.encrypt import AdaptiveEncryptionModel
from crypto_gltf.encrypt.deprecit.adaptive_v1.types import AdaptiveCipherParamsV1, Key
from crypto_gltf.io.plaintext.plnm import PlnM
from loguru import logger


class AdaptiveCryptoSystemV1(AdaptiveEncryptionModel, AdaptiveDecryptionModel):
    """Adaptive cipher system"""

    AAD: bool = True

    @staticmethod
    def encrypt(
        plnm: PlnM,
        meshes_cipher_params: AdaptiveCipherParamsV1,
        images_cipher_params: AdaptiveCipherParamsV1,
        key: Key,
        encrypt_images: bool = False,
    ) -> EncryptionResponse[PlnM, AAD_DATA, Key]:

        tic = time()
        # Get non-integer type arrays
        # i.e. vertices and colors
        meshes_float_arrs_idxs = np.array(
            [idx for idx, mesh in enumerate(plnm.meshes) if mesh.dtype.kind == "f"]
        )
        float_arrs = [
            plnm.meshes[i].astype(np.float32) for i in meshes_float_arrs_idxs
        ]  # convert to float32

        if key is None or key.k1 is None:
            key = Key(k1=secrets.token_bytes(32))

        encryption_response = AdaptiveEncryptionModel._encrypt(
            data=float_arrs, key=key, position=meshes_cipher_params.p
        )

        for idx, i in enumerate(meshes_float_arrs_idxs):
            plnm.meshes[i] = encryption_response.ciphertext[idx]

        logger.debug(f"Adaptive encryption took {time()-tic} seconds.")
        aad = AAD_DATA(
            aad=encryption_response.aad,
            encrypt_images=encrypt_images,
            meshes_params=meshes_cipher_params.__dict__,
            images_params=images_cipher_params.__dict__,
        )

        return EncryptionResponse[PlnM, AAD_DATA, Key](
            ciphertext=plnm, aad=aad, key=key
        )

    @staticmethod
    def decrypt(
        plnm: PlnM,
        key: Key,
        aad: AAD_DATA,
    ) -> PlnM:
        tic = time()
        # Get non-integer type arrays
        # i.e. vertices and colors
        meshes_float_arrs_idxs = np.array(
            [idx for idx, mesh in enumerate(plnm.meshes) if mesh.dtype.kind == "f"]
        )

        float_arrs = [plnm.meshes[i] for i in meshes_float_arrs_idxs]
        decrypted_data = AdaptiveDecryptionModel._decrypt(
            data=float_arrs,
            key=key,
            position=AdaptiveCipherParamsV1(**aad.meshes_params).p,
            aad=aad.aad,
        )
        for idx, i in enumerate(meshes_float_arrs_idxs):
            plnm.meshes[i] = decrypted_data[idx]

        logger.debug(f"Adaptive decryption took {time()-tic} seconds.")

        return plnm



import os
from urllib.parse import unquote


def uri_to_path(uri):
    uri = uri.replace("\\", "/")  # Some files come with \\ as dir separator
    uri = unquote(uri)
    return os.path.normpath(uri)



import numpy as np
from crypto_gltf.encrypt.deprecit.adaptive_v2.types import (
    AdaptiveCipherParamsV2,
    BlockSelection,
    SBlocks,
)


def combine_sblocks(sblocks_list: list[SBlocks]) -> SBlocks:
    """Combine sblock list into single sblock buffers"""
    s1, s2, s3 = bytearray(b""), bytearray(b""), bytearray(b"")
    for sblocks in sblocks_list:
        s1.extend(sblocks.s1)
        s2.extend(sblocks.s2)
        s3.extend(sblocks.s3)

    return SBlocks(s1=s1, s2=s2, s3=s3)


def get_sblocks(
    arr: np.ndarray,
    params: AdaptiveCipherParamsV2,
    to_get: BlockSelection,
) -> SBlocks:
    """Get byte at position for every element in float32 arr"""

    if arr.dtype != np.float32:
        raise Exception(f"Invalid array dtype {arr.dtype}, must be float32")

    bytes_arr = np.ndarray.tobytes(arr)  # flat bytes array
    p_buffer, q_buffer, r_buffer = bytearray(b""), bytearray(b""), bytearray(b"")
    for i in range(arr.size):  # iterate through all array elements
        if to_get.p:
            p_buffer.extend(bytes_arr[4 * i + params.p : 4 * i + params.p + 1])

        if to_get.q:
            q_buffer.extend(bytes_arr[4 * i + params.q : 4 * i + params.q + 1])

        if to_get.r:
            r_buffer.extend(bytes_arr[4 * i + params.r : 4 * i + params.r + 1])

    return SBlocks(s1=p_buffer, s2=q_buffer, s3=r_buffer)


def put_sblocks(
    arr: np.ndarray,
    params: AdaptiveCipherParamsV2,
    sblocks: SBlocks,
    offset: int,
    to_put: BlockSelection,
) -> np.ndarray:
    """Insert bytes at position in every element in float32 arr"""

    if arr.dtype != np.float32:
        raise Exception(f"Invalid array dtype {arr.dtype}, must be float32")

    bytes_arr = np.ndarray.tobytes(arr)  # flat bytes array
    for i in range(arr.size):  # iterate through all array elements
        if to_put.p:
            bytes_arr = (
                bytes_arr[: 4 * i + params.p]
                + sblocks.s1[offset + i : offset + i + 1]
                + bytes_arr[4 * i + params.p + 1 :]
            )
        if to_put.q:
            bytes_arr = (
                bytes_arr[: 4 * i + params.q]
                + sblocks.s2[offset + i : offset + i + 1]
                + bytes_arr[4 * i + params.q + 1 :]
            )
        if to_put.r:
            bytes_arr = (
                bytes_arr[: 4 * i + params.r]
                + sblocks.s3[offset + i : offset + i + 1]
                + bytes_arr[4 * i + params.r + 1 :]
            )

    return np.frombuffer(bytes_arr, dtype=np.float32).reshape(
        arr.shape
    )  # frombuffer returns a flat array, need to reshape



from enum import IntEnum

#
# Component Types
#
GLTF_COMPONENT_TYPE_BYTE = "BYTE"
GLTF_COMPONENT_TYPE_UNSIGNED_BYTE = "UNSIGNED_BYTE"
GLTF_COMPONENT_TYPE_SHORT = "SHORT"
GLTF_COMPONENT_TYPE_UNSIGNED_SHORT = "UNSIGNED_SHORT"
GLTF_COMPONENT_TYPE_UNSIGNED_INT = "UNSIGNED_INT"
GLTF_COMPONENT_TYPE_FLOAT = "FLOAT"


class ComponentType(IntEnum):
    Byte = 5120
    UnsignedByte = 5121
    Short = 5122
    UnsignedShort = 5123
    UnsignedInt = 5125
    Float = 5126

    @classmethod
    def to_type_code(cls, component_type):
        return {
            ComponentType.Byte: "b",
            ComponentType.UnsignedByte: "B",
            ComponentType.Short: "h",
            ComponentType.UnsignedShort: "H",
            ComponentType.UnsignedInt: "I",
            ComponentType.Float: "f",
        }[component_type]

    @classmethod
    def to_numpy_dtype(cls, component_type):
        import numpy as np

        return {
            ComponentType.Byte: np.int8,
            ComponentType.UnsignedByte: np.uint8,
            ComponentType.Short: np.int16,
            ComponentType.UnsignedShort: np.uint16,
            ComponentType.UnsignedInt: np.uint32,
            ComponentType.Float: np.float32,
        }[component_type]

    @classmethod
    def from_legacy_define(cls, type_define):
        return {
            GLTF_COMPONENT_TYPE_BYTE: ComponentType.Byte,
            GLTF_COMPONENT_TYPE_UNSIGNED_BYTE: ComponentType.UnsignedByte,
            GLTF_COMPONENT_TYPE_SHORT: ComponentType.Short,
            GLTF_COMPONENT_TYPE_UNSIGNED_SHORT: ComponentType.UnsignedShort,
            GLTF_COMPONENT_TYPE_UNSIGNED_INT: ComponentType.UnsignedInt,
            GLTF_COMPONENT_TYPE_FLOAT: ComponentType.Float,
        }[type_define]

    @classmethod
    def get_size(cls, component_type):
        return {
            ComponentType.Byte: 1,
            ComponentType.UnsignedByte: 1,
            ComponentType.Short: 2,
            ComponentType.UnsignedShort: 2,
            ComponentType.UnsignedInt: 4,
            ComponentType.Float: 4,
        }[component_type]


class DataType:
    Scalar = "SCALAR"
    Vec2 = "VEC2"
    Vec3 = "VEC3"
    Vec4 = "VEC4"
    Mat2 = "MAT2"
    Mat3 = "MAT3"
    Mat4 = "MAT4"

    def __new__(cls, *args, **kwargs):
        raise RuntimeError("{} should not be instantiated".format(cls.__name__))

    @classmethod
    def num_elements(cls, data_type):
        return {
            DataType.Scalar: 1,
            DataType.Vec2: 2,
            DataType.Vec3: 3,
            DataType.Vec4: 4,
            DataType.Mat2: 4,
            DataType.Mat3: 9,
            DataType.Mat4: 16,
        }[data_type]

    @classmethod
    def vec_type_from_num(cls, num_elems):
        if not (0 < num_elems < 5):
            raise ValueError("No vector type with {} elements".format(num_elems))
        return {
            1: DataType.Scalar,
            2: DataType.Vec2,
            3: DataType.Vec3,
            4: DataType.Vec4,
        }[num_elems]

    @classmethod
    def mat_type_from_num(cls, num_elems):
        if not (4 <= num_elems <= 16):
            raise ValueError("No matrix type with {} elements".format(num_elems))
        return {4: DataType.Mat2, 9: DataType.Mat3, 16: DataType.Mat4}[num_elems]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, Literal, NamedTuple, Type, TypeVar

import numpy as np
from pydantic import BaseModel, ConfigDict

JSONDict = dict[str, Any]

# FILETYPES


class Extension:
    """Extensions for filetypes"""

    ASSET = Literal[".off", ".glb", ".gltf"]
    IMAGE = Literal[".png", ".jpeg", ".jpg"]

    ANY_SUPPORTED = Literal[".off", ".png", ".jpeg", ".jpg"]


FileTypeName = Literal["ASSET", "IMAGE"]


class FileTypeConfig(NamedTuple):
    name: FileTypeName
    extensions: tuple[Extension.ANY_SUPPORTED, ...]


class FileType:
    ASSET = FileTypeConfig("ASSET", tuple(Extension.ASSET.__args__))
    IMAGE = FileTypeConfig("IMAGE", tuple(Extension.IMAGE.__args__))


# DATATYPES


@dataclass
class ArrayComposition3D:
    dim: int
    axis_restrictions: dict[Literal[0, 1, 2], int]
    dtype: type | None = None

    def __eq__(self, arr: np.ndarray) -> bool:
        """Check array is of composition composition"""
        shape, dtype = arr.shape, arr.dtype
        res = len(shape) == self.dim
        for k in self.axis_restrictions.keys():
            res = res and shape[k] == self.axis_restrictions[k]
        if self.dtype is not None:
            res = res and dtype == self.dtype
        return res

    @classmethod
    def from_data(cls, data: np.ndarray) -> ArrayComposition3D:
        """Initialise generalised composition from array"""
        return ArrayComposition3D(
            dim=len(data.shape), axis_restrictions={}, dtype=data.dtype
        )


class Composition:
    """Data type matrix composition"""

    FACES = ArrayComposition3D(dim=1, axis_restrictions={}, dtype=int)  # (n, 3)

    VERTS = ArrayComposition3D(
        dim=2, axis_restrictions={1: 3}, dtype=np.float32
    )  # (n, 3)

    RGB = ArrayComposition3D(
        dim=3, axis_restrictions={2: 3}, dtype=np.uint8
    )  # RGB (n, m, 3)

    GREYSCALE = ArrayComposition3D(
        dim=2, axis_restrictions={1: 2}
    )  # Black & white (n, 2)

    RGBA = ArrayComposition3D(dim=3, axis_restrictions={2: 4}, dtype=np.uint8)

    P = ArrayComposition3D(dim=2, axis_restrictions={}, dtype=np.uint8)

    AAD = ArrayComposition3D(dim=2, axis_restrictions={1: 3}, dtype=np.uint32)

    EMPTY = ArrayComposition3D(dim=1, axis_restrictions={0: 0})


DataTypeName = Literal[None, "GREYSCALE", "RGB", "MESH", "RGBA", "P"]


class PlnMDataType(NamedTuple):
    name: DataTypeName
    composition: ArrayComposition3D
    renderable: bool
    typed_name: str | None = None

    @classmethod
    def from_data(cls, data: np.ndarray) -> PlnMDataType:
        if Composition.FACES == data:
            return CombinedPlnMDataTypes.FACES
        elif Composition.VERTS == data:
            return CombinedPlnMDataTypes.VERTS
        elif Composition.GREYSCALE == data:
            return CombinedPlnMDataTypes.GREYSCALE
        elif Composition.RGB == data:
            return CombinedPlnMDataTypes.RGB
        elif Composition.RGBA == data:
            return CombinedPlnMDataTypes.RGBA
        elif Composition.P == data:
            return CombinedPlnMDataTypes.P
        else:
            composition = ArrayComposition3D.from_data(data)
            return cls(
                name=None,
                composition=composition,
                renderable=False,
                typed_name=None,
            )


class CombinedPlnMDataTypes:
    VERTS = PlnMDataType("MESH", Composition.VERTS, False)
    FACES = PlnMDataType("MESH", Composition.FACES, False)

    GREYSCALE = PlnMDataType("GREYSCALE", Composition.GREYSCALE, True, "L")
    RGB = PlnMDataType("RGB", Composition.RGB, True, "RGB")
    RGBA = PlnMDataType("RGBA", Composition.RGBA, True, "RGBA")
    P = PlnMDataType("P", Composition.P, True, "P")


# RESPONSE TYPES

T = TypeVar("T")
A = TypeVar("A")
K = TypeVar("K")


class EncryptionResponse(BaseModel, Generic[T, A, K]):
    """Generic response model for encryption model or system"""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    ciphertext: T
    aad: A
    key: K


class AAD_DATA(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    aad: np.ndarray
    encrypt_images: bool
    meshes_params: JSONDict
    images_params: JSONDict


# ENCRYPTION BASE TYPES


class BaseParams(BaseModel):
    """Base class for encryption model parameters"""

    p: int
    q: int
    r: int


class BaseKey(BaseModel):
    """Base class for encryption model keys"""

    k1: bytes | None = None
    k2: bytes | None = None
    k3: bytes | None = None



from __future__ import annotations

from time import sleep
from typing import Sequence

import numpy as np
from cellular_automaton import CellularAutomaton, EdgeRule, MooreNeighborhood

ALIVE = [1]
DEAD = [0]


class ConwaysCA(CellularAutomaton):
    """Cellular automaton with the evolution rules of conways game of life"""

    def __init__(self, dimension: tuple[int, int] = (100, 100)) -> None:
        super().__init__(
            dimension=dimension,
            neighborhood=MooreNeighborhood(
                EdgeRule.FIRST_AND_LAST_CELL_OF_DIMENSION_ARE_NEIGHBORS
            ),
        )

    def init_cell_state(self, __) -> Sequence:
        return [np.random.choice([0, 1])]

    def evolve_rule(self, last_cell_state, neighbors_last_states) -> Sequence:
        """Cellular automaton generation evolution rule"""
        new_cell_state = last_cell_state
        alive_neighbours = self.__count_alive_neighbours(neighbors_last_states)
        if last_cell_state == DEAD and alive_neighbours == 3:
            new_cell_state = ALIVE
        if last_cell_state == ALIVE and alive_neighbours < 2:
            new_cell_state = DEAD
        if last_cell_state == ALIVE and 1 < alive_neighbours < 4:
            new_cell_state = ALIVE
        if last_cell_state == ALIVE and alive_neighbours > 3:
            new_cell_state = DEAD
        return new_cell_state

    @staticmethod
    def __count_alive_neighbours(neighbours) -> int:
        """Count cell alive neighbours"""
        alive_neighbors = []
        for n in neighbours:
            if n == ALIVE:
                alive_neighbors.append(1)
        return len(alive_neighbors)


class ExpConwaysHCA(ConwaysCA):
    """Expanded hybrid cellular automata with the evolution rules of conways game of life"""

    expHCA: np.ndarray

    def __init__(
        self,
        dimension: tuple[int, int] = (100, 100),
    ) -> None:
        super().__init__(dimension=dimension)
        self.expHCA = np.reshape(
            self.ca_as_np_array, (1, self.dimension[0], self.dimension[1])
        )

    def insert_ps(self) -> None:
        """Insert a bogus bits generation"""
        np.append(
            self.expHCA,
            np.random.choice([0, 1], size=(self.dimension[0], self.dimension[1], 1)),
        )

    def diffuse(self) -> None:
        """Run a diffusion generation"""
        self.evolve()
        self.expHCA = np.vstack((self.expHCA, np.array([self.ca_as_np_array])))

    @property
    def ca_as_np_array(self) -> np.ndarray:
        """Get ca in np array format"""
        return np.array([cell.state for cell in self.get_cells().values()]).reshape(
            self.dimension
        )



from typing import ClassVar, Literal

import numpy as np
from crypto_gltf.data.types import AAD_DATA, BaseKey, BaseParams, EncryptionResponse
from crypto_gltf.io.plaintext.plnm import PlnM

CRYPTO_SYSTEMS = Literal["AdaptiveV1", "AdaptiveV2", "AdaptiveV3", "CA"]


class BaseCryptoSystem:
    """Base model for encryption/decryption system"""

    AAD: ClassVar[bool]  # true if the system uses aad data

    def __new__(cls, *args, **kwargs):
        raise RuntimeError(f"{cls} should not be instantiated")

    @staticmethod
    def encrypt(
        plnm: PlnM,
        meshes_cipher_params: BaseParams,
        images_cipher_params: BaseParams | None = None,
        key: BaseKey | None = None,
        encrypt_images: bool = False,
    ) -> EncryptionResponse[PlnM, AAD_DATA, BaseKey]:
        raise NotImplementedError()

    @staticmethod
    def decrypt(
        plnm: PlnM,
        key: BaseKey,
        aad: AAD_DATA,
    ) -> PlnM:
        raise NotImplementedError()



from __future__ import annotations

import numpy as np
from crypto_gltf.encrypt.deprecit.ca.base import CABaseModel, CACipherParams
from crypto_gltf.encrypt.deprecit.ca.ca import ExpConwaysHCA


class CADecryptionModel(CABaseModel):
    """Conways game of life cellular automata encryption model"""

    @staticmethod
    def decrypt(
        cipher_text: list[np.ndarray],
        expHCA: ExpConwaysHCA,
        cipher_params: CACipherParams,
    ) -> list[np.ndarray]:
        """Decrypt cipher text"""
        for i in range(cipher_params.gens - 1, -1, -1):

            if i in cipher_params.ps:
                continue
            else:
                cipher_text = CADecryptionModel.resolve_all_ct_matrices(
                    cipher_text=cipher_text,
                    expCA=expHCA.expHCA[i],
                    idl_points=cipher_params.idl_points,
                )

        return cipher_text

    @staticmethod
    def resolve_all_ct_matrices(
        cipher_text: list[np.ndarray],
        expCA: np.ndarray,
        idl_points: np.ndarray,
    ) -> list[np.ndarray]:
        """Inverse of confuse_all_pt_matrices: resolve all cipher text matrices.
        Conducts one resolve step"""
        resolved_data = [
            CADecryptionModel.resolve_single_ct_matrix(
                expCA=expCA, encrypted_text=mat, idl=idl
            )
            for mat, idl in zip(cipher_text, idl_points)
        ]
        return resolved_data

    @staticmethod
    def resolve_single_ct_matrix(
        expCA: np.ndarray, encrypted_text: np.ndarray, idl: tuple[int, int]
    ) -> np.ndarray:
        """Resolve a single ciphertext matrix one step.
        Split into 2D matrices and send to resolve_2d_matrix"""

        shape = encrypted_text.shape
        dims = len(shape)
        match dims:
            case 0:
                return encrypted_text
            case 1:
                return encrypted_text
            case 2:
                return CADecryptionModel.resolve_2d_matrix(
                    expCA=expCA, encrypted_text=encrypted_text, idl=idl
                )
            case 3:
                resolved_mat = np.empty(shape, dtype=encrypted_text.dtype)
                for i in range(shape[2]):
                    resolved_mat[:, :, i] = CADecryptionModel.resolve_2d_matrix(
                        expCA=expCA, encrypted_text=encrypted_text[:, :, i], idl=idl
                    )
                return resolved_mat
            case _:
                raise Exception(
                    f"Matrix has too many dimensions: {dims}, for CA encryptioon"
                )

    @staticmethod
    def resolve_2d_matrix(
        expCA: np.ndarray,
        encrypted_text: np.ndarray,
        idl: tuple[int, int],
    ) -> np.ndarray:
        """Resolve a single 2D ciphertext matrix one step"""
        if encrypted_text.size == 0:
            return encrypted_text
        flat_encrypted_text = encrypted_text.flatten()

        nonzero_elts_flat_idxs = CADecryptionModel.get_nonzero_idxs(
            expCA=expCA, mat=encrypted_text, idl=idl
        )
        zero_elts_flat_idxs = CADecryptionModel.get_zero_idxs(
            nonzero_elts_flat_idxs=nonzero_elts_flat_idxs, mat=encrypted_text
        )

        return np.concatenate(
            (
                np.take(flat_encrypted_text, nonzero_elts_flat_idxs),
                np.take(flat_encrypted_text, zero_elts_flat_idxs),
            )
        ).reshape(encrypted_text.shape)
